# Copyright (c) 2025 Airbyte, Inc., all rights reserved.
"""Secret hydration for Airbyte connectors."""

from collections.abc import Iterator
from copy import deepcopy
from functools import lru_cache
from typing import Any, cast

import requests
import yaml

from airbyte.constants import AIRBYTE_OFFLINE_MODE, SECRETS_HYDRATION_PREFIX
from airbyte.exceptions import PyAirbyteInternalError
from airbyte.secrets.util import get_secret


GLOBAL_MASK_KEYS_URL = "https://connectors.airbyte.com/files/registries/v0/specs_secrets_mask.yaml"


def _hydrate_recursive(
    config: dict,
) -> None:
    """Recursively hydrate secrets in the given configuration dictionary.

    Args:
        config (dict): The configuration dictionary to hydrate.
    """
    for key, value in config.items():
        if isinstance(value, dict):
            _hydrate_recursive(value)
        elif isinstance(value, list):
            for item in value:
                if isinstance(item, dict):
                    _hydrate_recursive(item)
        elif isinstance(value, str) and value.startswith(SECRETS_HYDRATION_PREFIX):
            # Assuming the value is a secret reference
            config[key] = get_secret(value)  # Replace with actual secret fetching logic


def hydrate_secrets(
    config: dict,
) -> dict:
    """Hydrate secrets in the given configuration dictionary.

    A new dictionary is created with secrets resolved, leaving the original unchanged.

    Args:
        config (dict): The configuration dictionary to hydrate.

    Returns:
        dict: The hydrated configuration dictionary with secrets resolved.
    """
    result: dict = deepcopy(config)
    _hydrate_recursive(result)
    return result


def deep_update(
    target: dict,
    source: dict,
) -> None:
    """Recursively update the target dictionary with values from the source dictionary.

    This is most often used to merge a non-secret configuration with a secret one, allowing
    users to keep non-secret config in source control while secrets are stored securely in
    a second configuration file or in environment variables.
    """
    for key, value in source.items():
        if isinstance(value, dict) and key in target and isinstance(target[key], dict):
            deep_update(target[key], value)
        else:
            target[key] = value


def _walk_dict(
    d: dict[str, Any],
    breadcrumb: list[str] | None = None,
) -> Iterator[tuple[list[str], Any]]:
    """Walk through a dictionary and yield paths to each leaf node.

    Yields tuples of (path, value) where path is a list of keys leading to the value.
    """
    breadcrumb = breadcrumb or []
    for key, val in d.items():
        new_path: list[str] = [*breadcrumb, key]
        if isinstance(val, dict):
            yield from _walk_dict(val, new_path)
        else:
            yield new_path, val


@lru_cache
def _get_global_secrets_mask() -> list[str]:
    """Get the list of properties to mask from the spec mask file."""
    if AIRBYTE_OFFLINE_MODE:
        # In offline mode, we cannot fetch the global mask keys.
        # We return an empty list to avoid masking any keys.
        return []

    response = requests.get(
        GLOBAL_MASK_KEYS_URL,
        allow_redirects=True,
    )
    if not response.ok:
        raise PyAirbyteInternalError(
            "Failed to parse spec mask.",
            log_text=response.content.decode("utf-8"),
        ) from None

    try:
        return cast("list[str]", yaml.safe_load(response.content)["properties"])

    except Exception as ex:
        raise PyAirbyteInternalError(
            "Failed to parse spec mask.",
            original_exception=ex,
        ) from None


def _get_connector_secrets_mask(
    spec_json_schema: dict[str, Any],
) -> list[str]:
    """Get the list of properties to mask from the connector spec."""
    result: list[str] = []
    for field_keys, field_value in _walk_dict(spec_json_schema):
        if isinstance(field_value, dict):
            is_secret: bool = any(
                (
                    field_value.get("writeOnly") is True,
                    field_value.get("format") == "password",
                    field_value.get("airbyte_secret") is True,
                )
            )
            if is_secret:
                result.append(field_keys[-1])

    return result


def detect_hardcoded_secrets(
    config: dict[str, Any],
    spec_json_schema: dict[str, Any] | None = None,
) -> list[list[str]]:
    """Detect hardcoded secrets in config.

    Apply this check before transmitting config over unsecured channels, to understand
    if the configuration contains hardcoded secrets.

    The recommended practice is to use environment variables or a secrets manager
    to store sensitive information, rather than hardcoding them in the configuration.

    To pass by reference, instead use the pattern `secret_reference::SECRET_NAME`,
    where `SECRET_NAME` is the name of the environment variable or a secret known
    by a registered secrets manager.
    """
    hardcoded_secrets: list[list[str]] = []
    secrets_mask: list[str] = (
        _get_global_secrets_mask()
        if spec_json_schema is None
        else _get_connector_secrets_mask(spec_json_schema)
    )

    for field_keys, field_value in _walk_dict(config):
        if (
            any(field_key in secrets_mask for field_key in field_keys)
            and isinstance(field_value, str)
            and not field_value.startswith(SECRETS_HYDRATION_PREFIX)
        ):
            hardcoded_secrets.append(field_keys)

    return hardcoded_secrets



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Connectivity to the connector catalog registry."""

from __future__ import annotations

import json
import logging
import os
import warnings
from copy import copy
from enum import Enum
from pathlib import Path
from typing import Any, cast

import requests
import yaml
from pydantic import BaseModel, Field
from typing_extensions import Self

from airbyte import exceptions as exc
from airbyte._registry_utils import fetch_registry_version_date, parse_changelog_html
from airbyte._util.meta import is_docker_installed
from airbyte.constants import AIRBYTE_OFFLINE_MODE
from airbyte.logs import warn_once
from airbyte.version import get_version


logger = logging.getLogger("airbyte")


__cache: dict[str, ConnectorMetadata] | None = None


_REGISTRY_ENV_VAR = "AIRBYTE_LOCAL_REGISTRY"
_REGISTRY_URL = "https://connectors.airbyte.com/files/registries/v0/oss_registry.json"

_PYTHON_LANGUAGE = "python"
_MANIFEST_ONLY_LANGUAGE = "manifest-only"

_PYTHON_LANGUAGE_TAG = f"language:{_PYTHON_LANGUAGE}"
_MANIFEST_ONLY_TAG = f"language:{_MANIFEST_ONLY_LANGUAGE}"

_DEFAULT_MANIFEST_URL = (
    "https://connectors.airbyte.com/files/metadata/airbyte/{source_name}/{version}/manifest.yaml"
)


class InstallType(str, Enum):
    """The type of installation for a connector."""

    YAML = "yaml"
    """Manifest-only connectors that can be run without Docker."""
    PYTHON = "python"
    """Python-based connectors available via PyPI."""
    DOCKER = "docker"
    """Docker-based connectors (returns all connectors for backward compatibility)."""
    JAVA = "java"
    """Java-based connectors."""

    INSTALLABLE = "installable"
    """Connectors installable in the current environment (environment-sensitive).

    Returns all connectors if Docker is installed, otherwise only Python and YAML.
    """
    ANY = "any"
    """All connectors in the registry (environment-independent)."""


class Language(str, Enum):
    """The language of a connector."""

    PYTHON = InstallType.PYTHON.value
    JAVA = InstallType.JAVA.value
    MANIFEST_ONLY = _MANIFEST_ONLY_LANGUAGE


class ConnectorMetadata(BaseModel):
    """Metadata for a connector."""

    name: str
    """Connector name. For example, "source-google-sheets"."""

    latest_available_version: str | None
    """The latest available version of the connector."""

    pypi_package_name: str | None
    """The name of the PyPI package for the connector, if it exists."""

    language: Language | None
    """The language of the connector."""

    install_types: set[InstallType]
    """The supported install types for the connector."""

    suggested_streams: list[str] | None = None
    """A list of suggested streams for the connector, if available."""

    @property
    def default_install_type(self) -> InstallType:
        """Return the default install type for the connector."""
        if self.language == Language.MANIFEST_ONLY and InstallType.YAML in self.install_types:
            return InstallType.YAML

        if InstallType.PYTHON in self.install_types:
            return InstallType.PYTHON

        # Else: Java or Docker
        return InstallType.DOCKER


def _get_registry_url() -> str:
    if _REGISTRY_ENV_VAR in os.environ:
        return str(os.environ.get(_REGISTRY_ENV_VAR))

    return _REGISTRY_URL


def _is_registry_disabled(url: str) -> bool:
    return url.upper() in {"0", "F", "FALSE"} or AIRBYTE_OFFLINE_MODE


def _registry_entry_to_connector_metadata(entry: dict) -> ConnectorMetadata:
    name = entry["dockerRepository"].replace("airbyte/", "")
    latest_version: str | None = entry.get("dockerImageTag")
    tags = entry.get("tags", [])
    language: Language | None = None

    if "language" in entry and entry["language"] is not None:
        try:
            language = Language(entry["language"])
        except Exception:
            warnings.warn(
                message=f"Invalid language for connector {name}: {entry['language']}",
                stacklevel=2,
            )
    if not language and _PYTHON_LANGUAGE_TAG in tags:
        language = Language.PYTHON
    if not language and _MANIFEST_ONLY_TAG in tags:
        language = Language.MANIFEST_ONLY

    remote_registries: dict = entry.get("remoteRegistries", {})
    pypi_registry: dict = remote_registries.get("pypi", {})
    pypi_package_name = cast(
        "str | None",
        pypi_registry.get("packageName", None),
    )
    pypi_enabled: bool = pypi_registry.get("enabled", False)
    install_types: set[InstallType] = {
        x
        for x in [
            InstallType.DOCKER,  # Always True
            InstallType.PYTHON if language == Language.PYTHON and pypi_enabled else None,
            InstallType.JAVA if language == Language.JAVA else None,
            InstallType.YAML if language == Language.MANIFEST_ONLY else None,
        ]
        if x
    }

    return ConnectorMetadata(
        name=name,
        latest_available_version=latest_version,
        pypi_package_name=pypi_package_name if pypi_enabled else None,
        language=language,
        install_types=install_types,
        suggested_streams=entry.get("suggestedStreams", {}).get("streams", None),
    )


def _get_registry_cache(
    *,
    force_refresh: bool = False,
) -> dict[str, ConnectorMetadata]:
    """Return the registry cache.

    Result is a mapping of connector name to ConnectorMetadata.
    """
    global __cache
    if __cache and not force_refresh:
        return __cache

    registry_url = _get_registry_url()

    if _is_registry_disabled(registry_url):
        return {}

    if registry_url.startswith("http"):
        response = requests.get(
            registry_url,
            headers={"User-Agent": f"PyAirbyte/{get_version()}"},
        )
        response.raise_for_status()
        data = response.json()
    else:
        # Assume local file
        with Path(registry_url).open(encoding="utf-8") as f:
            data = json.load(f)

    new_cache: dict[str, ConnectorMetadata] = {}

    for connector in data["sources"]:
        connector_metadata = _registry_entry_to_connector_metadata(connector)
        new_cache[connector_metadata.name] = connector_metadata

    for connector in data["destinations"]:
        connector_metadata = _registry_entry_to_connector_metadata(connector)
        new_cache[connector_metadata.name] = connector_metadata

    if len(new_cache) == 0:
        # This isn't necessarily fatal, since users can bring their own
        # connector definitions.
        warn_once(
            message=f"Connector registry is empty: {registry_url}",
            with_stack=False,
        )

    __cache = new_cache
    return __cache


def get_connector_metadata(name: str) -> ConnectorMetadata | None:
    """Check the cache for the connector.

    If the cache is empty, populate by calling update_cache.
    """
    registry_url = _get_registry_url()

    if _is_registry_disabled(registry_url):
        return None

    cache = copy(_get_registry_cache())

    if not cache:
        raise exc.PyAirbyteInternalError(
            message="Connector registry could not be loaded.",
            context={
                "registry_url": _get_registry_url(),
            },
        )
    if name not in cache:
        raise exc.AirbyteConnectorNotRegisteredError(
            connector_name=name,
            context={
                "registry_url": _get_registry_url(),
                "available_connectors": get_available_connectors(),
            },
        )
    return cache[name]


def get_available_connectors(
    install_type: InstallType | str | None = InstallType.INSTALLABLE,
) -> list[str]:
    """Return a list of all available connectors.

    Connectors will be returned in alphabetical order, with the standard prefix "source-".

    Args:
        install_type: The type of installation for the connector.
            Defaults to `InstallType.INSTALLABLE`.
    """
    if install_type is None or install_type == InstallType.INSTALLABLE:
        # Filter for installable connectors (default behavior).
        if is_docker_installed():
            logger.info("Docker is detected. Returning all connectors.")
            return sorted(_get_registry_cache().keys())

        logger.info("Docker was not detected. Returning only Python and Manifest-only connectors.")
        return sorted(
            [
                connector_name
                for connector_name, conn_info in _get_registry_cache().items()
                if conn_info.language in {Language.PYTHON, Language.MANIFEST_ONLY}
            ]
        )

    if not isinstance(install_type, InstallType):
        install_type = InstallType(install_type)

    if install_type == InstallType.PYTHON:
        return sorted(
            connector_name
            for connector_name, conn_info in _get_registry_cache().items()
            if conn_info.pypi_package_name is not None
        )

    if install_type == InstallType.JAVA:
        warnings.warn(
            message="Java connectors are not yet supported.",
            stacklevel=2,
        )
        return sorted(
            connector_name
            for connector_name, conn_info in _get_registry_cache().items()
            if conn_info.language == Language.JAVA
        )

    if install_type in {InstallType.DOCKER, InstallType.ANY}:
        return sorted(_get_registry_cache().keys())

    if install_type == InstallType.YAML:
        return sorted(
            conn.name
            for conn in _get_registry_cache().values()
            if InstallType.YAML in conn.install_types
        )

    # pragma: no cover  # Should never be reached.
    raise exc.PyAirbyteInputError(
        message="Invalid install type.",
        context={
            "install_type": install_type,
        },
    )


class ConnectorVersionInfo(BaseModel):
    """Information about a specific connector version."""

    version: str
    release_date: str | None = None
    docker_image_url: str
    changelog_url: str
    pr_url: str | None = None
    pr_title: str | None = None
    parsing_errors: list[str] = Field(default_factory=list)


class ApiDocsUrl(BaseModel):
    """API documentation URL information."""

    title: str
    url: str
    source: str
    doc_type: str = Field(default="other", alias="type")
    requires_login: bool = Field(default=False, alias="requiresLogin")

    model_config = {"populate_by_name": True}

    @classmethod
    def from_manifest_dict(cls, manifest_data: dict[str, Any]) -> list[Self]:
        """Extract documentation URLs from parsed manifest data.

        Args:
            manifest_data: The parsed manifest.yaml data as a dictionary

        Returns:
            List of ApiDocsUrl objects extracted from the manifest
        """
        results: list[Self] = []

        data_section = manifest_data.get("data")
        if isinstance(data_section, dict):
            external_docs = data_section.get("externalDocumentationUrls")
            if isinstance(external_docs, list):
                results = [
                    cls(
                        title=doc["title"],
                        url=doc["url"],
                        source="data_external_docs",
                        doc_type=doc.get("type", "other"),
                        requires_login=doc.get("requiresLogin", False),
                    )
                    for doc in external_docs
                ]

        return results


def _manifest_url_for(connector_name: str) -> str:
    """Get the expected URL of the manifest.yaml file for a connector.

    Args:
        connector_name: The canonical connector name (e.g., "source-facebook-marketing")

    Returns:
        The URL to the connector's manifest.yaml file
    """
    return _DEFAULT_MANIFEST_URL.format(
        source_name=connector_name,
        version="latest",
    )


def _fetch_manifest_dict(url: str) -> dict[str, Any]:
    """Fetch and parse a manifest.yaml file from a URL.

    Args:
        url: The URL to fetch the manifest from

    Returns:
        The parsed manifest data as a dictionary, or empty dict if manifest not found (404)

    Raises:
        HTTPError: If the request fails with a non-404 status code
    """
    http_not_found = 404

    response = requests.get(url, timeout=10)
    if response.status_code == http_not_found:
        return {}

    response.raise_for_status()
    return yaml.safe_load(response.text) or {}


def _extract_docs_from_registry(connector_name: str) -> list[ApiDocsUrl]:
    """Extract documentation URLs from connector registry metadata.

    Args:
        connector_name: The canonical connector name (e.g., "source-facebook-marketing")

    Returns:
        List of ApiDocsUrl objects extracted from the registry
    """
    registry_url = _get_registry_url()
    response = requests.get(registry_url, timeout=10)
    response.raise_for_status()
    registry_data = response.json()

    connector_list = registry_data.get("sources", []) + registry_data.get("destinations", [])
    connector_entry = None
    for entry in connector_list:
        if entry.get("dockerRepository", "").endswith(f"/{connector_name}"):
            connector_entry = entry
            break

    docs_urls = []

    if connector_entry and "documentationUrl" in connector_entry:
        docs_urls.append(
            ApiDocsUrl(
                title="Airbyte Documentation",
                url=connector_entry["documentationUrl"],
                source="registry",
            )
        )

    if connector_entry and "externalDocumentationUrls" in connector_entry:
        external_docs = connector_entry["externalDocumentationUrls"]
        if isinstance(external_docs, list):
            docs_urls.extend(
                [
                    ApiDocsUrl(
                        title=doc["title"],
                        url=doc["url"],
                        source="registry_external_docs",
                        doc_type=doc.get("type", "other"),
                        requires_login=doc.get("requiresLogin", False),
                    )
                    for doc in external_docs
                ]
            )

    return docs_urls


def get_connector_api_docs_urls(connector_name: str) -> list[ApiDocsUrl]:
    """Get API documentation URLs for a connector.

    This function retrieves documentation URLs for a connector's upstream API from multiple sources:
    - Registry metadata (documentationUrl, externalDocumentationUrls)
    - Connector manifest.yaml file (data.externalDocumentationUrls)

    Args:
        connector_name: The canonical connector name (e.g., "source-facebook-marketing")

    Returns:
        List of ApiDocsUrl objects with documentation URLs, deduplicated by URL.

    Raises:
        AirbyteConnectorNotRegisteredError: If the connector is not found in the registry.
    """
    if connector_name not in get_available_connectors(InstallType.ANY):
        raise exc.AirbyteConnectorNotRegisteredError(
            connector_name=connector_name,
            context={
                "registry_url": _get_registry_url(),
                "available_connectors": get_available_connectors(InstallType.ANY),
            },
        )

    docs_urls: list[ApiDocsUrl] = []

    registry_urls = _extract_docs_from_registry(connector_name)
    docs_urls.extend(registry_urls)

    manifest_url = _manifest_url_for(connector_name)
    manifest_data = _fetch_manifest_dict(manifest_url)
    manifest_urls = ApiDocsUrl.from_manifest_dict(manifest_data)
    docs_urls.extend(manifest_urls)

    seen_urls = set()
    unique_docs_urls = []
    for doc_url in docs_urls:
        if doc_url.url not in seen_urls:
            seen_urls.add(doc_url.url)
            unique_docs_urls.append(doc_url)

    return unique_docs_urls


def get_connector_version_history(
    connector_name: str,
    *,
    num_versions_to_validate: int = 5,
    timeout: int = 30,
) -> list[ConnectorVersionInfo]:
    """Get version history for a connector.

    This function retrieves the version history for a connector by:
    1. Scraping the changelog HTML from docs.airbyte.com
    2. Parsing version information including PR URLs and titles
    3. Overriding release dates for the most recent N versions with accurate
       registry data

    Args:
        connector_name: Name of the connector (e.g., 'source-faker', 'destination-postgres')
        num_versions_to_validate: Number of most recent versions to override with
            registry release dates for accuracy. Defaults to 5.
        timeout: Timeout in seconds for the changelog fetch. Defaults to 30.

    Returns:
        List of ConnectorVersionInfo objects, sorted by most recent first.

    Raises:
        AirbyteConnectorNotRegisteredError: If the connector is not found in the registry.

    Example:
        >>> versions = get_connector_version_history("source-faker", num_versions_to_validate=3)
        >>> for v in versions[:5]:
        ...     print(f"{v.version}: {v.release_date}")
    """
    if connector_name not in get_available_connectors(InstallType.ANY):
        raise exc.AirbyteConnectorNotRegisteredError(
            connector_name=connector_name,
            context={
                "registry_url": _get_registry_url(),
                "available_connectors": get_available_connectors(InstallType.ANY),
            },
        )

    connector_type = "sources" if connector_name.startswith("source-") else "destinations"
    connector_short_name = connector_name.replace("source-", "").replace("destination-", "")

    changelog_url = f"https://docs.airbyte.com/integrations/{connector_type}/{connector_short_name}"

    try:
        response = requests.get(
            changelog_url,
            headers={"User-Agent": f"PyAirbyte/{get_version()}"},
            timeout=timeout,
        )
        response.raise_for_status()
        html_content = response.text
    except requests.exceptions.RequestException as e:
        logger.warning(f"Failed to fetch changelog for {connector_name}: {e}")
        return []

    version_dicts = parse_changelog_html(html_content, connector_name)

    if not version_dicts:
        logger.warning(f"No versions found in changelog for {connector_name}")
        return []

    versions = [ConnectorVersionInfo(**version_dict) for version_dict in version_dicts]

    for version_info in versions[:num_versions_to_validate]:
        registry_date = fetch_registry_version_date(connector_name, version_info.version)
        if registry_date:
            version_info.release_date = registry_date
            logger.debug(
                f"Updated release date for {connector_name} v{version_info.version} "
                f"from registry: {registry_date}"
            )

    return versions



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Fixtures for integration tests."""

from __future__ import annotations

from contextlib import suppress
from typing import Any, Generator

import airbyte as ab
import pytest
from airbyte._util import meta, text_util
from airbyte._util.temp_files import as_temp_files
from airbyte.caches.base import CacheBase
from airbyte.caches.bigquery import BigQueryCache
from airbyte.caches.motherduck import MotherDuckCache
from airbyte.caches.snowflake import SnowflakeCache
from airbyte.destinations.base import Destination
from airbyte.secrets import GoogleGSMSecretManager, SecretHandle
from sqlalchemy import create_engine, text


AIRBYTE_INTERNAL_GCP_PROJECT = "dataline-integration-testing"


@pytest.fixture(scope="session")
def ci_secret_manager() -> GoogleGSMSecretManager:
    secret = ab.get_secret("GCP_GSM_CREDENTIALS")
    if not secret or secret.is_empty():
        pytest.skip("GCP_GSM_CREDENTIALS secret not found.")

    return GoogleGSMSecretManager(
        project=AIRBYTE_INTERNAL_GCP_PROJECT,
        credentials_json=ab.get_secret("GCP_GSM_CREDENTIALS"),
    )


def get_connector_config(self, connector_name: str, index: int = 0) -> dict | None:
    """Retrieve the connector configuration from GSM."""
    gsm_secrets_manager = GoogleGSMSecretManager(
        project=AIRBYTE_INTERNAL_GCP_PROJECT,
        credentials_json=ab.get_secret("GCP_GSM_CREDENTIALS"),
    )
    first_secret: SecretHandle = next(
        gsm_secrets_manager.fetch_connector_secrets(
            connector_name=connector_name,
        ),
        None,
    )

    print(f"Found '{connector_name}' credential secret '${first_secret.secret_name}'.")
    return first_secret.get_value().parse_json()


@pytest.fixture(scope="session")
def motherduck_secrets(ci_secret_manager: GoogleGSMSecretManager) -> dict:
    return ci_secret_manager.get_secret(
        "SECRET_DESTINATION_DUCKDB__MOTHERDUCK__CREDS",
    ).parse_json()


@pytest.fixture
def new_motherduck_cache(
    motherduck_secrets,
) -> MotherDuckCache:
    return MotherDuckCache(
        database="my_db",  # TODO: Use a dedicated DB for testing
        schema_name=f"test_deleteme_{text_util.generate_random_suffix()}",
        api_key=motherduck_secrets["motherduck_api_key"],
    )


@pytest.fixture
def new_motherduck_destination(
    motherduck_secrets,
) -> Destination:
    return ab.get_destination(
        "destination-motherduck",
        config=motherduck_secrets,
        install_if_missing=False,
    )


@pytest.fixture(scope="session")
def new_snowflake_destination_config(ci_secret_manager: GoogleGSMSecretManager) -> dict:
    config = ci_secret_manager.get_secret(
        "AIRBYTE_LIB_SNOWFLAKE_CREDS",
    ).parse_json()
    config["schema"] = f"test_deleteme_{text_util.generate_random_suffix()}"
    return config


@pytest.fixture
def new_snowflake_cache(
    new_snowflake_destination_config: dict[str, Any],
) -> Generator[SnowflakeCache, Any, None]:
    cache = SnowflakeCache(
        account=new_snowflake_destination_config["account"],
        username=new_snowflake_destination_config["username"],
        password=new_snowflake_destination_config["password"],
        database=new_snowflake_destination_config["database"],
        warehouse=new_snowflake_destination_config["warehouse"],
        role=new_snowflake_destination_config["role"],
        schema_name=new_snowflake_destination_config["schema"],
    )
    sqlalchemy_url = cache.get_sql_alchemy_url()

    yield cache

    engine = create_engine(
        sqlalchemy_url,
        future=True,
    )
    with engine.connect() as connection:
        connection.execute(
            text(f"DROP SCHEMA IF EXISTS {cache.schema_name}"),
        )


@pytest.fixture
def new_snowflake_destination(
    new_snowflake_destination_config: dict[str, Any],
) -> Destination:
    dest_config = new_snowflake_destination_config.copy()
    _ = dest_config.pop("destinationType", None)
    return ab.get_destination(
        "destination-snowflake",
        config=dest_config,
        install_if_missing=False,
    )


@pytest.fixture
def new_bigquery_destination_config(
    ci_secret_manager: GoogleGSMSecretManager,
) -> dict[str, Any]:
    dest_bigquery_config = ci_secret_manager.get_secret(
        "SECRET_DESTINATION-BIGQUERY_CREDENTIALS__CREDS"
    ).parse_json()
    dest_bigquery_config["dataset_id"] = (
        f"test_deleteme_{text_util.generate_random_suffix()}"
    )
    return dest_bigquery_config


@pytest.fixture
def new_bigquery_destination(
    new_bigquery_destination_config: dict[str, Any],
) -> Destination:
    dest_config = new_bigquery_destination_config.copy()
    _ = dest_config.pop("destinationType", None)
    return ab.get_destination(
        "destination-bigquery",
        config=dest_config,
        install_if_missing=False,
    )


@pytest.fixture
def new_bigquery_cache(
    new_bigquery_destination_config: dict[str, Any],
) -> Generator[BigQueryCache, Any, None]:
    credentials_json = new_bigquery_destination_config["credentials_json"]
    with as_temp_files([credentials_json]) as (credentials_path,):
        cache = BigQueryCache(
            credentials_path=credentials_path,
            project_name=new_bigquery_destination_config["project_id"],
            dataset_name=new_bigquery_destination_config["dataset_id"],
        )
        yield cache

        url = cache.get_sql_alchemy_url()
        engine = create_engine(
            url,
            future=True,
        )
        with suppress(Exception):
            with engine.begin() as connection:
                connection.execute(text(f"DROP SCHEMA IF EXISTS {cache.schema_name}"))


@pytest.fixture(scope="function")
def new_generic_cache(request) -> CacheBase:
    """This is a placeholder fixture that will be overridden by pytest_generate_tests()."""
    return request.getfixturevalue(request.param)


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """Override default pytest behavior, parameterizing our tests based on the available cache types.

    This is useful for running the same tests with different cache types, to ensure that the tests
    can pass across all cache types.
    """
    all_cache_type_fixtures: dict[str, str] = {
        # Ordered by priority (fastest first)
        "DuckDB": "new_duckdb_cache",
        "Postgres": "new_postgres_cache",
        "BigQuery": "new_bigquery_cache",
        "Snowflake": "new_snowflake_cache",
    }
    if meta.is_windows():
        # Postgres tests require Linux containers
        all_cache_type_fixtures.pop("Postgres")

    if "new_generic_cache" in metafunc.fixturenames:
        metafunc.parametrize(
            "new_generic_cache",
            all_cache_type_fixtures.values(),
            ids=all_cache_type_fixtures.keys(),
            indirect=True,
            scope="function",
        )



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Experimental features which may change.

> **NOTE:**
> The following "experimental" features are now "stable" and can be accessed directly from the
`airbyte.get_source()` method:
> - Docker sources, using the `docker_image` argument.
> - Yaml sources, using the `source_manifest` argument.

## About Experimental Features

Experimental features may change without notice between minor versions of PyAirbyte. Although rare,
they may also be entirely removed or refactored in future versions of PyAirbyte. Experimental
features may also be less stable than other features, and may not be as well-tested.

You can help improve this product by reporting issues and providing feedback for improvements in our
[GitHub issue tracker](https://github.com/airbytehq/pyairbyte/issues).
"""

from __future__ import annotations

from airbyte.sources.util import get_source


__all__ = [
    "get_source",
]



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A Postgres implementation of the PyAirbyte cache.

## Usage Example

```python
from airbyte as ab
from airbyte.caches import PostgresCache

cache = PostgresCache(
    host="myhost",
    port=5432,
    username="myusername",
    password=ab.get_secret("POSTGRES_PASSWORD"),
    database="mydatabase",
)
```
"""

from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from airbyte_api.models import DestinationPostgres

from airbyte._processors.sql.postgres import PostgresConfig, PostgresSqlProcessor
from airbyte.caches.base import CacheBase
from airbyte.destinations._translate_cache_to_dest import (
    postgres_cache_to_destination_configuration,
)


if TYPE_CHECKING:
    from airbyte.shared.sql_processor import SqlProcessorBase


class PostgresCache(PostgresConfig, CacheBase):
    """Configuration for the Postgres cache.

    Also inherits config from the JsonlWriter, which is responsible for writing files to disk.
    """

    _sql_processor_class: ClassVar[type[SqlProcessorBase]] = PostgresSqlProcessor

    paired_destination_name: ClassVar[str | None] = "destination-bigquery"
    paired_destination_config_class: ClassVar[type | None] = DestinationPostgres

    @property
    def paired_destination_config(self) -> DestinationPostgres:
        """Return a dictionary of destination configuration values."""
        return postgres_cache_to_destination_configuration(cache=self)

    def clone_as_cloud_destination_config(self) -> DestinationPostgres:
        """Return a DestinationPostgres instance with the same configuration."""
        return DestinationPostgres(
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database,
        )


# Expose the Cache class and also the Config class.
__all__ = [
    "PostgresCache",
    "PostgresConfig",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Airbyte Cloud MCP operations."""

# Note: Deferred type evaluation must be avoided due to FastMCP/Pydantic needing
# types to be available at import time for tool registration.
import contextlib
import logging
from typing import Annotated, Any, Literal

import requests
from fastmcp import FastMCP
from fastmcp_extensions import mcp_tool, register_mcp_tools
from pydantic import BaseModel, Field

from airbyte import exceptions as exc
from airbyte._util.meta import is_docker_installed
from airbyte.mcp._arg_resolvers import resolve_list_of_strings
from airbyte.registry import (
    _DEFAULT_MANIFEST_URL,
    ApiDocsUrl,
    ConnectorMetadata,
    ConnectorVersionInfo,
    InstallType,
    get_available_connectors,
    get_connector_api_docs_urls,
    get_connector_metadata,
)
from airbyte.registry import get_connector_version_history as _get_connector_version_history
from airbyte.sources.util import get_source


logger = logging.getLogger("airbyte.mcp")


@mcp_tool(
    read_only=True,
    idempotent=True,
)
def list_connectors(
    keyword_filter: Annotated[
        str | None,
        Field(
            description="Filter connectors by keyword.",
            default=None,
        ),
    ],
    connector_type_filter: Annotated[
        Literal["source", "destination"] | None,
        Field(
            description="Filter connectors by type ('source' or 'destination').",
            default=None,
        ),
    ],
    install_types: Annotated[
        Literal["java", "python", "yaml", "docker"]
        | list[Literal["java", "python", "yaml", "docker"]]
        | None,
        Field(
            description=(
                """
                Filter connectors by install type.
                These are not mutually exclusive:
                - "python": Connectors that can be installed as Python packages.
                - "yaml": Connectors that can be installed simply via YAML download.
                    These connectors are the fastest to install and run, as they do not require any
                    additional dependencies.
                - "java": Connectors that can only be installed via Java. Since PyAirbyte does not
                    currently ship with a JVM, these connectors will be run via Docker instead.
                    In environments where Docker is not available, these connectors may not be
                    runnable.
                - "docker": Connectors that can be installed via Docker. Note that all connectors
                    can be run in Docker, so this filter should generally return the same results as
                    not specifying a filter.
                If no install types are specified, all connectors will be returned.
                """
            ),
            default=None,
        ),
    ],
) -> list[str]:
    """List available Airbyte connectors with optional filtering.

    Returns:
        List of connector names.
    """
    # Start with the full list of known connectors (all support Docker):
    connectors: list[str] = get_available_connectors(install_type=InstallType.ANY)

    install_types_list: list[str] | None = resolve_list_of_strings(
        install_types,  # type: ignore[arg-type]  # Type check doesn't understand literal is str
    )

    if install_types_list:
        # If install_types is provided, filter connectors based on the specified install types.
        connectors = [
            connector
            for connector in connectors
            if any(
                connector in get_available_connectors(install_type=install_type)
                for install_type in install_types_list
            )
        ]

    if keyword_filter:
        # Filter connectors by keyword, case-insensitive.
        connectors = [
            connector for connector in connectors if keyword_filter.lower() in connector.lower()
        ]

    if connector_type_filter:
        # Filter connectors by type ('source' or 'destination').
        # This assumes connector names are prefixed with 'source-' or 'destination-'.
        connectors = [
            connector
            for connector in connectors
            if connector.startswith(f"{connector_type_filter}-")
        ]

    return sorted(connectors)


class ConnectorInfo(BaseModel):
    """@private Class to hold connector information."""

    connector_name: str
    connector_metadata: ConnectorMetadata | None = None
    documentation_url: str | None = None
    config_spec_jsonschema: dict | None = None
    manifest_url: str | None = None


@mcp_tool(
    read_only=True,
    idempotent=True,
)
def get_connector_info(
    connector_name: Annotated[
        str,
        Field(description="The name of the connector to get information for."),
    ],
) -> ConnectorInfo | Literal["Connector not found."]:
    """Get the documentation URL for a connector."""
    if connector_name not in get_available_connectors():
        return "Connector not found."

    connector = get_source(
        connector_name,
        docker_image=is_docker_installed() or False,
        install_if_missing=False,  # Defer to avoid failing entirely if it can't be installed.
    )

    connector_metadata: ConnectorMetadata | None = None
    with contextlib.suppress(Exception):
        connector_metadata = get_connector_metadata(connector_name)

    config_spec_jsonschema: dict[str, Any] | None = None
    with contextlib.suppress(Exception):
        # This requires running the connector. Install it if it isn't already installed.
        connector.install()
        config_spec_jsonschema = connector.config_spec

    manifest_url = _DEFAULT_MANIFEST_URL.format(
        source_name=connector_name,
        version="latest",
    )

    return ConnectorInfo(
        connector_name=connector.name,
        connector_metadata=connector_metadata,
        documentation_url=connector.docs_url,
        config_spec_jsonschema=config_spec_jsonschema,
        manifest_url=manifest_url,
    )


@mcp_tool(
    read_only=True,
    idempotent=True,
)
def get_api_docs_urls(
    connector_name: Annotated[
        str,
        Field(
            description=(
                "The canonical connector name "
                "(e.g., 'source-facebook-marketing', 'destination-snowflake')"
            )
        ),
    ],
) -> list[ApiDocsUrl] | Literal["Connector not found."]:
    """Get API documentation URLs for a connector.

    This tool retrieves documentation URLs for a connector's upstream API from multiple sources:
    - Registry metadata (documentationUrl, externalDocumentationUrls)
    - Connector manifest.yaml file (data.externalDocumentationUrls)
    """
    try:
        return get_connector_api_docs_urls(connector_name)
    except exc.AirbyteConnectorNotRegisteredError:
        return "Connector not found."


@mcp_tool(
    read_only=True,
    idempotent=True,
)
def get_connector_version_history(
    connector_name: Annotated[
        str,
        Field(
            description="The name of the connector (e.g., 'source-faker', 'destination-postgres')"
        ),
    ],
    num_versions_to_validate: Annotated[
        int,
        Field(
            description=(
                "Number of most recent versions to validate with registry data for accurate "
                "release dates. Defaults to 5."
            ),
            default=5,
        ),
    ] = 5,
    limit: Annotated[
        int | None,
        Field(
            description=(
                "DEPRECATED: Use num_versions_to_validate instead. "
                "Maximum number of versions to return (most recent first). "
                "If specified, only the first N versions will be returned."
            ),
            default=None,
        ),
    ] = None,
) -> list[ConnectorVersionInfo] | Literal["Connector not found.", "Failed to fetch changelog."]:
    """Get version history for a connector.

    This tool retrieves the version history for a connector, including:
    - Version number
    - Release date (from changelog, with registry override for recent versions)
    - DockerHub URL for the version
    - Changelog URL
    - PR URL and title (scraped from changelog)

    For the most recent N versions (default 5), release dates are fetched from the
    registry for accuracy. For older versions, changelog dates are used.

    Returns:
        List of version information, sorted by most recent first.
    """
    try:
        versions = _get_connector_version_history(
            connector_name=connector_name,
            num_versions_to_validate=num_versions_to_validate,
        )
    except exc.AirbyteConnectorNotRegisteredError:
        return "Connector not found."
    except requests.exceptions.RequestException:
        logger.exception(f"Failed to fetch changelog for {connector_name}")
        return "Failed to fetch changelog."
    else:
        if limit is not None and limit > 0:
            return versions[:limit]
        return versions


def register_registry_tools(app: FastMCP) -> None:
    """Register registry tools with the FastMCP app.

    Args:
        app: FastMCP application instance
    """
    register_mcp_tools(app, mcp_module=__name__)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Unit tests for connection state models and helper functions."""

from __future__ import annotations

import pytest

from airbyte.cloud._connection_state import (
    ConnectionStateResponse,
    GlobalState,
    StreamDescriptor,
    StreamState,
    _get_stream_list,
    _match_stream,
)


@pytest.mark.parametrize(
    "stream_name,stream_namespace,descriptor_name,descriptor_namespace,expected",
    [
        pytest.param("users", None, "users", None, True, id="name_match_no_namespace"),
        pytest.param(
            "users", None, "users", "public", False, id="no_wildcard_namespace"
        ),
        pytest.param("users", "", "users", None, True, id="empty_str_matches_none"),
        pytest.param("users", None, "users", "", True, id="none_matches_empty_str"),
        pytest.param(
            "users", "public", "users", "public", True, id="name_and_namespace_match"
        ),
        pytest.param(
            "users", "public", "users", "private", False, id="namespace_mismatch"
        ),
        pytest.param("orders", None, "users", None, False, id="name_mismatch"),
        pytest.param("users", "public", "users", None, False, id="namespace_vs_none"),
    ],
)
def test_match_stream(
    stream_name: str,
    stream_namespace: str | None,
    descriptor_name: str,
    descriptor_namespace: str | None,
    expected: bool,
) -> None:
    stream = StreamState(
        stream_descriptor=StreamDescriptor(
            name=descriptor_name,
            namespace=descriptor_namespace,
        ),
        stream_state={"cursor": "2024-01-01"},
    )
    assert _match_stream(stream, stream_name, stream_namespace) == expected


@pytest.mark.parametrize(
    "state_type,has_stream_state,has_global_state,expected_count",
    [
        pytest.param("stream", True, False, 2, id="stream_type_with_streams"),
        pytest.param("global", False, True, 2, id="global_type_with_streams"),
        pytest.param("legacy", False, False, 0, id="legacy_type_empty"),
        pytest.param("not_set", False, False, 0, id="not_set_empty"),
        pytest.param("stream", False, False, 0, id="stream_type_no_streams"),
    ],
)
def test_get_stream_list(
    state_type: str,
    has_stream_state: bool,
    has_global_state: bool,
    expected_count: int,
) -> None:
    streams = [
        StreamState(
            stream_descriptor=StreamDescriptor(name="users"),
            stream_state={"cursor": "2024-01-01"},
        ),
        StreamState(
            stream_descriptor=StreamDescriptor(name="orders"),
            stream_state={"cursor": "2024-06-01"},
        ),
    ]
    state = ConnectionStateResponse(
        state_type=state_type,
        connection_id="test-conn-id",
        stream_state=streams if has_stream_state else None,
        global_state=GlobalState(stream_states=streams) if has_global_state else None,
    )
    result = _get_stream_list(state)
    assert len(result) == expected_count


def test_connection_state_response_str_stream_type() -> None:
    state = ConnectionStateResponse(
        state_type="stream",
        connection_id="abc-123",
        stream_state=[
            StreamState(
                stream_descriptor=StreamDescriptor(name="users"),
                stream_state={"cursor": "2024-01-01"},
            ),
        ],
    )
    assert "abc-123" in str(state)
    assert "stateType=stream" in str(state)
    assert "streams=1" in str(state)


def test_connection_state_response_str_global_type() -> None:
    state = ConnectionStateResponse(
        state_type="global",
        connection_id="abc-456",
        global_state=GlobalState(
            shared_state={"cdc": True},
            stream_states=[
                StreamState(
                    stream_descriptor=StreamDescriptor(name="users"),
                    stream_state={"cursor": "2024-01-01"},
                ),
                StreamState(
                    stream_descriptor=StreamDescriptor(name="orders"),
                    stream_state={"cursor": "2024-06-01"},
                ),
            ],
        ),
    )
    assert "abc-456" in str(state)
    assert "stateType=global" in str(state)
    assert "streams=2" in str(state)


def test_connection_state_response_from_camel_case() -> None:
    raw_api_response = {
        "stateType": "stream",
        "connectionId": "test-conn-id",
        "streamState": [
            {
                "streamDescriptor": {"name": "users", "namespace": "public"},
                "streamState": {"cursor": "2024-01-01"},
            },
        ],
    }
    state = ConnectionStateResponse(**raw_api_response)
    assert state.state_type == "stream"
    assert state.connection_id == "test-conn-id"
    assert state.stream_state is not None
    assert len(state.stream_state) == 1
    assert state.stream_state[0].stream_descriptor.name == "users"
    assert state.stream_state[0].stream_descriptor.namespace == "public"
    assert state.stream_state[0].stream_state == {"cursor": "2024-01-01"}


def test_connection_state_response_global_from_camel_case() -> None:
    raw_api_response = {
        "stateType": "global",
        "connectionId": "test-conn-id",
        "globalState": {
            "sharedState": {"cdc_offset": "12345"},
            "streamStates": [
                {
                    "streamDescriptor": {"name": "users"},
                    "streamState": {"cursor": "2024-01-01"},
                },
            ],
        },
    }
    state = ConnectionStateResponse(**raw_api_response)
    assert state.state_type == "global"
    assert state.global_state is not None
    assert state.global_state.shared_state == {"cdc_offset": "12345"}
    assert len(state.global_state.stream_states) == 1
    assert state.global_state.stream_states[0].stream_descriptor.name == "users"



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""State backend implementation."""

from __future__ import annotations

import abc
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from airbyte_protocol.models import (
        AirbyteStreamState,
    )

    from airbyte.shared.sql_processor import SqlConfig
    from airbyte.shared.state_providers import StateProviderBase
    from airbyte.shared.state_writers import StateWriterBase


class StateBackendBase(abc.ABC):
    """A class which manages the stream state for data synced.

    The backend is responsible for storing and retrieving the state of streams. It generates
    `StateProvider` objects, which are paired to a specific source and table prefix.
    """

    _sql_config: SqlConfig

    def __init__(self) -> None:
        """Initialize the state manager with a static catalog state."""
        self._state_artifacts: list[AirbyteStreamState] | None = None

    @abc.abstractmethod
    def get_state_provider(
        self,
        source_name: str,
        table_prefix: str,
        *,
        refresh: bool = True,
        destination_name: str | None = None,
    ) -> StateProviderBase:
        """Return the state provider."""
        ...

    @abc.abstractmethod
    def get_state_writer(
        self,
        source_name: str,
        destination_name: str | None = None,
    ) -> StateWriterBase:
        """Return a state writer for a named source.

        The same table prefix of the backend will be used for the state writer.
        """
        ...

    def _initialize_backend(
        self,
        *,
        force_refresh: bool = False,
    ) -> None:
        """Do any needed initialization, for instance to load state artifacts from the cache.

        By default, this method does nothing. Base classes may override this method to load state
        artifacts or perform other initialization tasks.
        """
        _ = force_refresh  # Unused
        pass



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""PyAirbyte Records module.

## Understanding record handling in PyAirbyte

PyAirbyte models record handling after Airbyte's "Destination V2" ("Dv2") record handling. This
includes the below implementation details.

### Field Name Normalization

1. PyAirbyte normalizes top-level record keys to lowercase, replacing spaces and hyphens with
   underscores.
2. PyAirbyte does not normalize nested keys on sub-properties.

For example, the following record:

```json
{

    "My-Field": "value",
    "Nested": {
        "MySubField": "value"
    }
}
```

Would be normalized to:

```json
{
    "my_field": "value",
    "nested": {
        "MySubField": "value"
    }
}
```

### Table Name Normalization

Similar to column handling, PyAirbyte normalizes table names to the lowercase version of the stream
name and may remove or normalize special characters.

### Airbyte-Managed Metadata Columns

PyAirbyte adds the following columns to every record:

- `ab_raw_id`: A unique identifier for the record.
- `ab_extracted_at`: The time the record was extracted.
- `ab_meta`: A dictionary of extra metadata about the record.

The names of these columns are included in the `airbyte.constants` module for programmatic
reference.

## Schema Evolution

PyAirbyte supports a very basic form of schema evolution:

1. Columns are always auto-added to cache tables whenever newly arriving properties are detected
   as not present in the cache table.
2. Column types will not be modified or expanded to fit changed types in the source catalog.
   - If column types change, we recommend user to manually alter the column types.
3. At any time, users can run a full sync with a `WriteStrategy` of 'replace'. This will create a
   fresh table from scratch and then swap the old and new tables after table sync is complete.

---

"""

from __future__ import annotations

from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any

from uuid_extensions import uuid7str

from airbyte._util.name_normalizers import LowerCaseNormalizer, NameNormalizerBase
from airbyte.constants import (
    AB_EXTRACTED_AT_COLUMN,
    AB_INTERNAL_COLUMNS,
    AB_META_COLUMN,
    AB_RAW_ID_COLUMN,
)


if TYPE_CHECKING:
    from collections.abc import Iterator

    from airbyte_protocol.models import (
        AirbyteRecordMessage,
    )


class StreamRecordHandler:
    """A class to handle processing of StreamRecords.

    This is a long-lived object that can be used to process multiple StreamRecords, which
    saves on memory and processing time by reusing the same object for all records of the same type.
    """

    def __init__(
        self,
        *,
        json_schema: dict,
        normalizer: type[NameNormalizerBase] = LowerCaseNormalizer,
        normalize_keys: bool = False,
        prune_extra_fields: bool,
    ) -> None:
        """Initialize the dictionary with the given data.

        Args:
            json_schema: The JSON Schema definition for this stream.
            normalizer: The normalizer to use when normalizing keys. If not provided, the
                LowerCaseNormalizer will be used.
            normalize_keys: If `True`, the keys will be normalized using the given normalizer.
            prune_extra_fields: If `True`, unexpected fields will be removed.
        """
        self._expected_keys: list[str] = list(json_schema.get("properties", {}).keys())
        self._normalizer: type[NameNormalizerBase] = normalizer
        self._normalize_keys: bool = normalize_keys
        self.prune_extra_fields: bool = prune_extra_fields

        self.index_keys: list[str] = [
            self._normalizer.normalize(key) if self._normalize_keys else key
            for key in self._expected_keys
        ]
        self.normalized_keys: list[str] = [
            self._normalizer.normalize(key) for key in self._expected_keys
        ]
        self.quick_lookup: dict[str, str]

        for internal_col in AB_INTERNAL_COLUMNS:
            if internal_col not in self._expected_keys:
                self._expected_keys.append(internal_col)

        # Store a lookup from normalized keys to pretty cased (originally cased) keys.
        self._pretty_case_lookup: dict[str, str] = {
            self._normalizer.normalize(pretty_case.lower()): pretty_case
            for pretty_case in self._expected_keys
        }
        # Store a map from all key versions (normalized and pretty-cased) to their normalized
        # version.
        self.quick_lookup = {
            key: self._normalizer.normalize(key)
            if self._normalize_keys
            else self.to_display_case(key)
            for key in set(self._expected_keys) | set(self._pretty_case_lookup.values())
        }

    def to_display_case(self, key: str) -> str:
        """Return the original case of the key.

        If the key is not found in the pretty case lookup, return the key provided.
        """
        return self._pretty_case_lookup.get(self._normalizer.normalize(key), key)

    def to_index_case(self, key: str) -> str:
        """Return the internal case of the key.

        If `normalize_keys` is True, returns the normalized key.
        Otherwise, return the original case ("pretty case") of the key.
        """
        try:
            return self.quick_lookup[key]
        except KeyError:
            result = (
                self._normalizer.normalize(key)
                if self._normalize_keys
                else self.to_display_case(key)
            )
            self.quick_lookup[key] = result
            return result


class StreamRecord(dict[str, Any]):
    """The StreamRecord class is a case-aware, case-insensitive dictionary implementation.

    It has these behaviors:
    - When a key is retrieved, deleted, or checked for existence, it is always checked in a
      case-insensitive manner.
    - The original case is stored in a separate dictionary, so that the original case can be
      retrieved when needed.
    - Because it is subclassed from `dict`, the `StreamRecord` class can be passed as a normal
      Python dictionary.
    - In addition to the properties of the stream's records, the dictionary also stores the Airbyte
      metadata columns: `_airbyte_raw_id`, `_airbyte_extracted_at`, and `_airbyte_meta`.

    This behavior mirrors how a case-aware, case-insensitive SQL database would handle column
    references.

    There are two ways this class can store keys internally:
    - If normalize_keys is True, the keys are normalized using the given normalizer.
    - If normalize_keys is False, the original case of the keys is stored.

    In regards to missing values, the dictionary accepts an 'expected_keys' input. When set, the
    dictionary will be initialized with the given keys. If a key is not found in the input data, it
    will be initialized with a value of None. When provided, the 'expected_keys' input will also
    determine the original case of the keys.
    """

    def __init__(
        self,
        from_dict: dict,
        *,
        stream_record_handler: StreamRecordHandler,
        with_internal_columns: bool = True,
        extracted_at: datetime | None = None,
    ) -> None:
        """Initialize the dictionary with the given data.

        Args:
            from_dict: The dictionary to initialize the StreamRecord with.
            stream_record_handler: The StreamRecordHandler to use for processing the record.
            with_internal_columns: If `True`, the internal columns will be added to the record.
            extracted_at: The time the record was extracted. If not provided, the current time will
                be used.
        """
        self._stream_handler: StreamRecordHandler = stream_record_handler

        # Start by initializing all values to None
        self.update(dict.fromkeys(stream_record_handler.index_keys))

        # Update the dictionary with the given data
        if self._stream_handler.prune_extra_fields:
            self.update(
                {
                    self._stream_handler.to_index_case(k): v
                    for k, v in from_dict.items()
                    if self._stream_handler.to_index_case(k) in self._stream_handler.index_keys
                }
            )
        else:
            self.update({self._stream_handler.to_index_case(k): v for k, v in from_dict.items()})

        if with_internal_columns:
            self.update(
                {
                    AB_RAW_ID_COLUMN: uuid7str(),
                    AB_EXTRACTED_AT_COLUMN: extracted_at or datetime.now(timezone.utc),
                    AB_META_COLUMN: {},
                }
            )

    @classmethod
    def from_record_message(
        cls,
        record_message: AirbyteRecordMessage,
        *,
        stream_record_handler: StreamRecordHandler,
    ) -> StreamRecord:
        """Return a StreamRecord from a RecordMessage."""
        data_dict: dict[str, Any] = record_message.data.copy()
        return cls(
            from_dict=data_dict,
            stream_record_handler=stream_record_handler,
            with_internal_columns=True,
            extracted_at=datetime.fromtimestamp(record_message.emitted_at / 1000, tz=timezone.utc),
        )

    def __getitem__(self, key: str) -> Any:  # noqa: ANN401
        """Return the item with the given key."""
        try:
            return super().__getitem__(key)
        except KeyError:
            return super().__getitem__(self._stream_handler.to_index_case(key))

    def __setitem__(self, key: str, value: Any) -> None:  # noqa: ANN401
        """Set the item with the given key to the given value."""
        index_case_key = self._stream_handler.to_index_case(key)
        if (
            self._stream_handler.prune_extra_fields
            and index_case_key not in self._stream_handler.index_keys
        ):
            return

        super().__setitem__(index_case_key, value)

    def __delitem__(self, key: str) -> None:
        """Delete the item with the given key."""
        try:
            super().__delitem__(key)
        except KeyError:
            index_case_key = self._stream_handler.to_index_case(key)
            if super().__contains__(index_case_key):
                super().__delitem__(index_case_key)
                return
        else:
            # No failure. Key was deleted.
            return

        raise KeyError(key)

    def __contains__(self, key: object) -> bool:
        """Return whether the dictionary contains the given key."""
        assert isinstance(key, str), "Key must be a string."
        return super().__contains__(key) or super().__contains__(
            self._stream_handler.to_index_case(key)
        )

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the keys of the dictionary."""
        return iter(super().__iter__())

    def __len__(self) -> int:
        """Return the number of items in the dictionary."""
        return super().__len__()

    def __eq__(self, other: object) -> bool:
        """Return whether the StreamRecord is equal to the given dict or StreamRecord object."""
        if isinstance(other, StreamRecord):
            return dict(self) == dict(other)

        if isinstance(other, dict):
            return {k.lower(): v for k, v in self.items()} == {
                k.lower(): v for k, v in other.items()
            }
        return False

    def __hash__(self) -> int:  # type: ignore [override]  # Doesn't match superclass (dict)
        """Return the hash of the dictionary with keys sorted."""
        items = [(k, v) for k, v in self.items() if not isinstance(v, dict)]
        return hash(tuple(sorted(items)))



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
from __future__ import annotations

from pathlib import Path
from typing import Optional
import pytest_mock
from airbyte.caches.snowflake import SnowflakeSqlProcessor, SnowflakeConfig
from airbyte_protocol.models import ConfiguredAirbyteCatalog
from airbyte.secrets.base import SecretString
from airbyte.shared.catalog_providers import CatalogProvider


def test_snowflake_cache_config_data_retention_time_in_days(
    mocker: pytest_mock.MockFixture,
):
    expected_cmd = """
        CREATE TABLE airbyte_raw."table_name" (
            col_name type
        )
        DATA_RETENTION_TIME_IN_DAYS = 1
        """

    def _execute_sql(cmd):
        global actual_cmd
        actual_cmd = cmd

    mocker.patch.object(SnowflakeSqlProcessor, "_execute_sql", side_effect=_execute_sql)
    config = _build_mocked_snowflake_processor(mocker, data_retention_time_in_days=1)
    config._create_table(table_name="table_name", column_definition_str="col_name type")

    assert actual_cmd == expected_cmd


def test_snowflake_cache_config_no_data_retention_time_in_days(
    mocker: pytest_mock.MockFixture,
):
    expected_cmd = """
        CREATE TABLE airbyte_raw."table_name" (
            col_name type
        )
        \n        """

    def _execute_sql(cmd):
        global actual_cmd
        actual_cmd = cmd

    mocker.patch.object(SnowflakeSqlProcessor, "_execute_sql", side_effect=_execute_sql)
    config = _build_mocked_snowflake_processor(mocker)
    config._create_table(table_name="table_name", column_definition_str="col_name type")

    assert actual_cmd == expected_cmd


def _build_mocked_snowflake_processor(
    mocker: pytest_mock.MockFixture, data_retention_time_in_days: Optional[int] = None
):
    sql_config = SnowflakeConfig(
        account="foo",
        username="foo",
        password=SecretString("foo"),
        warehouse="foo",
        database="foo",
        role="foo",
        data_retention_time_in_days=data_retention_time_in_days,
    )

    mocker.patch.object(
        SnowflakeSqlProcessor, "_ensure_schema_exists", return_value=None
    )
    return SnowflakeSqlProcessor(
        catalog_provider=CatalogProvider(ConfiguredAirbyteCatalog(streams=[])),
        temp_dir=Path(),
        temp_file_cleanup=True,
        sql_config=sql_config,
    )



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""PyAirbyte classes and methods for interacting with the Airbyte Cloud API.

By overriding `api_root`, you can use this module to interact with self-managed Airbyte instances,
both OSS and Enterprise.

## Usage Examples

Get a new workspace object and deploy a source to it:

```python
import airbyte as ab
from airbyte import cloud

workspace = cloud.CloudWorkspace(
    workspace_id="...",
    client_id="...",
    client_secret="...",
)

# Deploy a source to the workspace
source = ab.get_source("source-faker", config={"count": 100})
deployed_source = workspace.deploy_source(
    name="test-source",
    source=source,
)

# Run a check on the deployed source and raise an exception if the check fails
check_result = deployed_source.check(raise_on_error=True)

# Permanently delete the newly-created source
workspace.permanently_delete_source(deployed_source)
```
"""

from __future__ import annotations

from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal, overload

import yaml

from airbyte import exceptions as exc
from airbyte._util import api_util, text_util
from airbyte._util.api_util import get_web_url_root
from airbyte.cloud.auth import (
    resolve_cloud_api_url,
    resolve_cloud_bearer_token,
    resolve_cloud_client_id,
    resolve_cloud_client_secret,
    resolve_cloud_workspace_id,
)
from airbyte.cloud.client_config import CloudClientConfig
from airbyte.cloud.connections import CloudConnection
from airbyte.cloud.connectors import (
    CloudDestination,
    CloudSource,
    CustomCloudSourceDefinition,
)
from airbyte.destinations.base import Destination
from airbyte.exceptions import AirbyteError
from airbyte.secrets.base import SecretString


if TYPE_CHECKING:
    from collections.abc import Callable

    from airbyte.sources.base import Source


class CloudOrganization:
    """Information about an organization in Airbyte Cloud.

    This class provides lazy loading of organization attributes including billing status.
    It is typically created via CloudWorkspace.get_organization().
    """

    def __init__(
        self,
        organization_id: str,
        organization_name: str | None = None,
        email: str | None = None,
        *,
        api_root: str = api_util.CLOUD_API_ROOT,
        client_id: SecretString | None = None,
        client_secret: SecretString | None = None,
        bearer_token: SecretString | None = None,
    ) -> None:
        """Initialize a CloudOrganization.

        Args:
            organization_id: The organization ID.
            organization_name: Display name of the organization.
            email: Email associated with the organization.
            api_root: The API root URL.
            client_id: OAuth client ID for authentication.
            client_secret: OAuth client secret for authentication.
            bearer_token: Bearer token for authentication (alternative to client credentials).
        """
        self.organization_id = organization_id
        """The organization ID."""

        self._organization_name = organization_name
        """Display name of the organization."""

        self._email = email
        """Email associated with the organization."""

        self._api_root = api_root
        self._client_id = client_id
        self._client_secret = client_secret
        self._bearer_token = bearer_token

        # Cached organization info (billing, etc.)
        self._organization_info: dict[str, Any] | None = None
        # Flag to remember if fetching organization info failed (e.g., permission issues)
        self._organization_info_fetch_failed: bool = False

    def _fetch_organization_info(self, *, force_refresh: bool = False) -> dict[str, Any]:
        """Fetch and cache organization info including billing status.

        If fetching fails (e.g., due to permission issues), the failure is cached and
        subsequent calls will return an empty dict without retrying.

        Args:
            force_refresh: If True, always fetch from the API even if cached.

        Returns:
            Dictionary containing organization info including billing data.
            Returns empty dict if fetching failed or is not permitted.
        """
        # Reset failure flag if force_refresh is requested
        if force_refresh:
            self._organization_info_fetch_failed = False

        # If we already know fetching failed, return empty dict without retrying
        if self._organization_info_fetch_failed:
            return {}

        if not force_refresh and self._organization_info is not None:
            return self._organization_info

        try:
            self._organization_info = api_util.get_organization_info(
                organization_id=self.organization_id,
                api_root=self._api_root,
                client_id=self._client_id,
                client_secret=self._client_secret,
                bearer_token=self._bearer_token,
            )
        except Exception:
            # Cache the failure so we don't retry on subsequent property accesses
            self._organization_info_fetch_failed = True
            return {}
        else:
            return self._organization_info

    @property
    def organization_name(self) -> str | None:
        """Display name of the organization."""
        if self._organization_name is not None:
            return self._organization_name
        # Try to fetch from API if not set (returns empty dict on failure)
        info = self._fetch_organization_info()
        return info.get("organizationName")

    @property
    def email(self) -> str | None:
        """Email associated with the organization."""
        if self._email is not None:
            return self._email
        # Try to fetch from API if not set (returns empty dict on failure)
        info = self._fetch_organization_info()
        return info.get("email")

    @property
    def payment_status(self) -> str | None:
        """Payment status of the organization.

        Possible values: 'uninitialized', 'okay', 'grace_period', 'disabled', 'locked', 'manual'.
        When 'disabled', syncs are blocked due to unpaid invoices.
        Returns None if billing info is not available (e.g., due to permission issues).
        """
        info = self._fetch_organization_info()
        return (info.get("billing") or {}).get("paymentStatus")

    @property
    def subscription_status(self) -> str | None:
        """Subscription status of the organization.

        Possible values: 'pre_subscription', 'subscribed', 'unsubscribed'.
        Returns None if billing info is not available (e.g., due to permission issues).
        """
        info = self._fetch_organization_info()
        return (info.get("billing") or {}).get("subscriptionStatus")

    @property
    def is_account_locked(self) -> bool:
        """Whether the account is locked due to billing issues.

        Returns True if payment_status is 'disabled'/'locked' or subscription_status is
        'unsubscribed'. Defaults to False unless we have affirmative evidence of a locked state.
        """
        return api_util.is_account_locked(self.payment_status, self.subscription_status)


@dataclass
class CloudWorkspace:
    """A remote workspace on the Airbyte Cloud.

    By overriding `api_root`, you can use this class to interact with self-managed Airbyte
    instances, both OSS and Enterprise.

    Two authentication methods are supported (mutually exclusive):
    1. OAuth2 client credentials (client_id + client_secret)
    2. Bearer token authentication

    Example with client credentials:
        ```python
        workspace = CloudWorkspace(
            workspace_id="...",
            client_id="...",
            client_secret="...",
        )
        ```

    Example with bearer token:
        ```python
        workspace = CloudWorkspace(
            workspace_id="...",
            bearer_token="...",
        )
        ```
    """

    workspace_id: str
    client_id: SecretString | None = None
    client_secret: SecretString | None = None
    api_root: str = api_util.CLOUD_API_ROOT
    bearer_token: SecretString | None = None

    # Internal credentials object (set in __post_init__, excluded from __init__)
    _credentials: CloudClientConfig | None = field(default=None, init=False, repr=False)

    def __post_init__(self) -> None:
        """Validate and initialize credentials."""
        # Wrap secrets in SecretString if provided
        if self.client_id is not None:
            self.client_id = SecretString(self.client_id)
        if self.client_secret is not None:
            self.client_secret = SecretString(self.client_secret)
        if self.bearer_token is not None:
            self.bearer_token = SecretString(self.bearer_token)

        # Create internal CloudClientConfig object (validates mutual exclusivity)
        self._credentials = CloudClientConfig(
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
            api_root=self.api_root,
        )

    @classmethod
    def from_env(
        cls,
        workspace_id: str | None = None,
        *,
        api_root: str | None = None,
    ) -> CloudWorkspace:
        """Create a CloudWorkspace using credentials from environment variables.

        This factory method resolves credentials from environment variables,
        providing a convenient way to create a workspace without explicitly
        passing credentials.

        Two authentication methods are supported (mutually exclusive):
        1. Bearer token (checked first)
        2. OAuth2 client credentials (fallback)

        Environment variables used:
            - `AIRBYTE_CLOUD_BEARER_TOKEN`: Bearer token (alternative to client credentials).
            - `AIRBYTE_CLOUD_CLIENT_ID`: OAuth client ID (for client credentials flow).
            - `AIRBYTE_CLOUD_CLIENT_SECRET`: OAuth client secret (for client credentials flow).
            - `AIRBYTE_CLOUD_WORKSPACE_ID`: The workspace ID (if not passed as argument).
            - `AIRBYTE_CLOUD_API_URL`: Optional. The API root URL (defaults to Airbyte Cloud).

        Args:
            workspace_id: The workspace ID. If not provided, will be resolved from
                the `AIRBYTE_CLOUD_WORKSPACE_ID` environment variable.
            api_root: The API root URL. If not provided, will be resolved from
                the `AIRBYTE_CLOUD_API_URL` environment variable, or default to
                the Airbyte Cloud API.

        Returns:
            A CloudWorkspace instance configured with credentials from the environment.

        Raises:
            PyAirbyteSecretNotFoundError: If required credentials are not found in
                the environment.

        Example:
            ```python
            # With workspace_id from environment
            workspace = CloudWorkspace.from_env()

            # With explicit workspace_id
            workspace = CloudWorkspace.from_env(workspace_id="your-workspace-id")
            ```
        """
        resolved_api_root = resolve_cloud_api_url(api_root)

        # Try bearer token first
        bearer_token = resolve_cloud_bearer_token()
        if bearer_token:
            return cls(
                workspace_id=resolve_cloud_workspace_id(workspace_id),
                bearer_token=bearer_token,
                api_root=resolved_api_root,
            )

        # Fall back to client credentials
        return cls(
            workspace_id=resolve_cloud_workspace_id(workspace_id),
            client_id=resolve_cloud_client_id(),
            client_secret=resolve_cloud_client_secret(),
            api_root=resolved_api_root,
        )

    @property
    def workspace_url(self) -> str | None:
        """The web URL of the workspace."""
        return f"{get_web_url_root(self.api_root)}/workspaces/{self.workspace_id}"

    @cached_property
    def _organization_info(self) -> dict[str, Any]:
        """Fetch and cache organization info for this workspace.

        Uses the Config API endpoint for an efficient O(1) lookup.
        This is an internal method; use get_organization() for public access.
        """
        return api_util.get_workspace_organization_info(
            workspace_id=self.workspace_id,
            api_root=self.api_root,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )

    @overload
    def get_organization(self) -> CloudOrganization: ...

    @overload
    def get_organization(
        self,
        *,
        raise_on_error: Literal[True],
    ) -> CloudOrganization: ...

    @overload
    def get_organization(
        self,
        *,
        raise_on_error: Literal[False],
    ) -> CloudOrganization | None: ...

    def get_organization(
        self,
        *,
        raise_on_error: bool = True,
    ) -> CloudOrganization | None:
        """Get the organization this workspace belongs to.

        Fetching organization info requires ORGANIZATION_READER permissions on the organization,
        which may not be available with workspace-scoped credentials.

        Args:
            raise_on_error: If True (default), raises AirbyteError on permission or API errors.
                If False, returns None instead of raising.

        Returns:
            CloudOrganization object with organization_id and organization_name,
            or None if raise_on_error=False and an error occurred.

        Raises:
            AirbyteError: If raise_on_error=True and the organization info cannot be fetched
                (e.g., due to insufficient permissions or missing data).
        """
        try:
            info = self._organization_info
        except (AirbyteError, NotImplementedError):
            if raise_on_error:
                raise
            return None

        organization_id = info.get("organizationId")
        organization_name = info.get("organizationName")

        # Validate that both organization_id and organization_name are non-null and non-empty
        if not organization_id or not organization_name:
            if raise_on_error:
                raise AirbyteError(
                    message="Organization info is incomplete.",
                    context={
                        "organization_id": organization_id,
                        "organization_name": organization_name,
                    },
                )
            return None

        return CloudOrganization(
            organization_id=organization_id,
            organization_name=organization_name,
            api_root=self.api_root,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )

    # Test connection and creds

    def connect(self) -> None:
        """Check that the workspace is reachable and raise an exception otherwise.

        Note: It is not necessary to call this method before calling other operations. It
              serves primarily as a simple check to ensure that the workspace is reachable
              and credentials are correct.
        """
        _ = api_util.get_workspace(
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        print(f"Successfully connected to workspace: {self.workspace_url}")

    # Get sources, destinations, and connections

    def get_connection(
        self,
        connection_id: str,
    ) -> CloudConnection:
        """Get a connection by ID.

        This method does not fetch data from the API. It returns a `CloudConnection` object,
        which will be loaded lazily as needed.
        """
        return CloudConnection(
            workspace=self,
            connection_id=connection_id,
        )

    def get_source(
        self,
        source_id: str,
    ) -> CloudSource:
        """Get a source by ID.

        This method does not fetch data from the API. It returns a `CloudSource` object,
        which will be loaded lazily as needed.
        """
        return CloudSource(
            workspace=self,
            connector_id=source_id,
        )

    def get_destination(
        self,
        destination_id: str,
    ) -> CloudDestination:
        """Get a destination by ID.

        This method does not fetch data from the API. It returns a `CloudDestination` object,
        which will be loaded lazily as needed.
        """
        return CloudDestination(
            workspace=self,
            connector_id=destination_id,
        )

    # Deploy sources and destinations

    def deploy_source(
        self,
        name: str,
        source: Source,
        *,
        unique: bool = True,
        random_name_suffix: bool = False,
    ) -> CloudSource:
        """Deploy a source to the workspace.

        Returns the newly deployed source.

        Args:
            name: The name to use when deploying.
            source: The source object to deploy.
            unique: Whether to require a unique name. If `True`, duplicate names
                are not allowed. Defaults to `True`.
            random_name_suffix: Whether to append a random suffix to the name.
        """
        source_config_dict = source._hydrated_config.copy()  # noqa: SLF001 (non-public API)
        source_config_dict["sourceType"] = source.name.replace("source-", "")

        if random_name_suffix:
            name += f" (ID: {text_util.generate_random_suffix()})"

        if unique:
            existing = self.list_sources(name=name)
            if existing:
                raise exc.AirbyteDuplicateResourcesError(
                    resource_type="source",
                    resource_name=name,
                )

        deployed_source = api_util.create_source(
            name=name,
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            config=source_config_dict,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        return CloudSource(
            workspace=self,
            connector_id=deployed_source.source_id,
        )

    def deploy_destination(
        self,
        name: str,
        destination: Destination | dict[str, Any],
        *,
        unique: bool = True,
        random_name_suffix: bool = False,
    ) -> CloudDestination:
        """Deploy a destination to the workspace.

        Returns the newly deployed destination ID.

        Args:
            name: The name to use when deploying.
            destination: The destination to deploy. Can be a local Airbyte `Destination` object or a
                dictionary of configuration values.
            unique: Whether to require a unique name. If `True`, duplicate names
                are not allowed. Defaults to `True`.
            random_name_suffix: Whether to append a random suffix to the name.
        """
        if isinstance(destination, Destination):
            destination_conf_dict = destination._hydrated_config.copy()  # noqa: SLF001 (non-public API)
            destination_conf_dict["destinationType"] = destination.name.replace("destination-", "")
            # raise ValueError(destination_conf_dict)
        else:
            destination_conf_dict = destination.copy()
            if "destinationType" not in destination_conf_dict:
                raise exc.PyAirbyteInputError(
                    message="Missing `destinationType` in configuration dictionary.",
                )

        if random_name_suffix:
            name += f" (ID: {text_util.generate_random_suffix()})"

        if unique:
            existing = self.list_destinations(name=name)
            if existing:
                raise exc.AirbyteDuplicateResourcesError(
                    resource_type="destination",
                    resource_name=name,
                )

        deployed_destination = api_util.create_destination(
            name=name,
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            config=destination_conf_dict,  # Wants a dataclass but accepts dict
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        return CloudDestination(
            workspace=self,
            connector_id=deployed_destination.destination_id,
        )

    def permanently_delete_source(
        self,
        source: str | CloudSource,
        *,
        safe_mode: bool = True,
    ) -> None:
        """Delete a source from the workspace.

        You can pass either the source ID `str` or a deployed `Source` object.

        Args:
            source: The source ID or CloudSource object to delete
            safe_mode: If True, requires the source name to contain "delete-me" or "deleteme"
                (case insensitive) to prevent accidental deletion. Defaults to True.
        """
        if not isinstance(source, (str, CloudSource)):
            raise exc.PyAirbyteInputError(
                message="Invalid source type.",
                input_value=type(source).__name__,
            )

        api_util.delete_source(
            source_id=source.connector_id if isinstance(source, CloudSource) else source,
            source_name=source.name if isinstance(source, CloudSource) else None,
            api_root=self.api_root,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
            safe_mode=safe_mode,
        )

    # Deploy and delete destinations

    def permanently_delete_destination(
        self,
        destination: str | CloudDestination,
        *,
        safe_mode: bool = True,
    ) -> None:
        """Delete a deployed destination from the workspace.

        You can pass either the `Cache` class or the deployed destination ID as a `str`.

        Args:
            destination: The destination ID or CloudDestination object to delete
            safe_mode: If True, requires the destination name to contain "delete-me" or "deleteme"
                (case insensitive) to prevent accidental deletion. Defaults to True.
        """
        if not isinstance(destination, (str, CloudDestination)):
            raise exc.PyAirbyteInputError(
                message="Invalid destination type.",
                input_value=type(destination).__name__,
            )

        api_util.delete_destination(
            destination_id=(
                destination if isinstance(destination, str) else destination.destination_id
            ),
            destination_name=(
                destination.name if isinstance(destination, CloudDestination) else None
            ),
            api_root=self.api_root,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
            safe_mode=safe_mode,
        )

    # Deploy and delete connections

    def deploy_connection(
        self,
        connection_name: str,
        *,
        source: CloudSource | str,
        selected_streams: list[str],
        destination: CloudDestination | str,
        table_prefix: str | None = None,
    ) -> CloudConnection:
        """Create a new connection between an already deployed source and destination.

        Returns the newly deployed connection object.

        Args:
            connection_name: The name of the connection.
            source: The deployed source. You can pass a source ID or a CloudSource object.
            destination: The deployed destination. You can pass a destination ID or a
                CloudDestination object.
            table_prefix: Optional. The table prefix to use when syncing to the destination.
            selected_streams: The selected stream names to sync within the connection.
        """
        if not selected_streams:
            raise exc.PyAirbyteInputError(
                guidance="You must provide `selected_streams` when creating a connection."
            )

        source_id: str = source if isinstance(source, str) else source.connector_id
        destination_id: str = (
            destination if isinstance(destination, str) else destination.connector_id
        )

        deployed_connection = api_util.create_connection(
            name=connection_name,
            source_id=source_id,
            destination_id=destination_id,
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            selected_stream_names=selected_streams,
            prefix=table_prefix or "",
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )

        return CloudConnection(
            workspace=self,
            connection_id=deployed_connection.connection_id,
            source=deployed_connection.source_id,
            destination=deployed_connection.destination_id,
        )

    def permanently_delete_connection(
        self,
        connection: str | CloudConnection,
        *,
        cascade_delete_source: bool = False,
        cascade_delete_destination: bool = False,
        safe_mode: bool = True,
    ) -> None:
        """Delete a deployed connection from the workspace.

        Args:
            connection: The connection ID or CloudConnection object to delete
            cascade_delete_source: If True, also delete the source after deleting the connection
            cascade_delete_destination: If True, also delete the destination after deleting
                the connection
            safe_mode: If True, requires the connection name to contain "delete-me" or "deleteme"
                (case insensitive) to prevent accidental deletion. Defaults to True. Also applies
                to cascade deletes.
        """
        if connection is None:
            raise ValueError("No connection ID provided.")

        if isinstance(connection, str):
            connection = CloudConnection(
                workspace=self,
                connection_id=connection,
            )

        api_util.delete_connection(
            connection_id=connection.connection_id,
            connection_name=connection.name,
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
            safe_mode=safe_mode,
        )

        if cascade_delete_source:
            self.permanently_delete_source(
                source=connection.source_id,
                safe_mode=safe_mode,
            )
        if cascade_delete_destination:
            self.permanently_delete_destination(
                destination=connection.destination_id,
                safe_mode=safe_mode,
            )

    # List sources, destinations, and connections

    def list_connections(
        self,
        name: str | None = None,
        *,
        name_filter: Callable | None = None,
    ) -> list[CloudConnection]:
        """List connections by name in the workspace.

        TODO: Add pagination support
        """
        connections = api_util.list_connections(
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            name=name,
            name_filter=name_filter,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        return [
            CloudConnection._from_connection_response(  # noqa: SLF001 (non-public API)
                workspace=self,
                connection_response=connection,
            )
            for connection in connections
            if name is None or connection.name == name
        ]

    def list_sources(
        self,
        name: str | None = None,
        *,
        name_filter: Callable | None = None,
    ) -> list[CloudSource]:
        """List all sources in the workspace.

        TODO: Add pagination support
        """
        sources = api_util.list_sources(
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            name=name,
            name_filter=name_filter,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        return [
            CloudSource._from_source_response(  # noqa: SLF001 (non-public API)
                workspace=self,
                source_response=source,
            )
            for source in sources
            if name is None or source.name == name
        ]

    def list_destinations(
        self,
        name: str | None = None,
        *,
        name_filter: Callable | None = None,
    ) -> list[CloudDestination]:
        """List all destinations in the workspace.

        TODO: Add pagination support
        """
        destinations = api_util.list_destinations(
            api_root=self.api_root,
            workspace_id=self.workspace_id,
            name=name,
            name_filter=name_filter,
            client_id=self.client_id,
            client_secret=self.client_secret,
            bearer_token=self.bearer_token,
        )
        return [
            CloudDestination._from_destination_response(  # noqa: SLF001 (non-public API)
                workspace=self,
                destination_response=destination,
            )
            for destination in destinations
            if name is None or destination.name == name
        ]

    def publish_custom_source_definition(
        self,
        name: str,
        *,
        manifest_yaml: dict[str, Any] | Path | str | None = None,
        docker_image: str | None = None,
        docker_tag: str | None = None,
        unique: bool = True,
        pre_validate: bool = True,
        testing_values: dict[str, Any] | None = None,
    ) -> CustomCloudSourceDefinition:
        """Publish a custom source connector definition.

        You must specify EITHER manifest_yaml (for YAML connectors) OR both docker_image
        and docker_tag (for Docker connectors), but not both.

        Args:
            name: Display name for the connector definition
            manifest_yaml: Low-code CDK manifest (dict, Path to YAML file, or YAML string)
            docker_image: Docker repository (e.g., 'airbyte/source-custom')
            docker_tag: Docker image tag (e.g., '1.0.0')
            unique: Whether to enforce name uniqueness
            pre_validate: Whether to validate manifest client-side (YAML only)
            testing_values: Optional configuration values to use for testing in the
                Connector Builder UI. If provided, these values are stored as the complete
                testing values object for the connector builder project (replaces any existing
                values), allowing immediate test read operations.

        Returns:
            CustomCloudSourceDefinition object representing the created definition

        Raises:
            PyAirbyteInputError: If both or neither of manifest_yaml and docker_image provided
            AirbyteDuplicateResourcesError: If unique=True and name already exists
        """
        is_yaml = manifest_yaml is not None
        is_docker = docker_image is not None

        if is_yaml == is_docker:
            raise exc.PyAirbyteInputError(
                message=(
                    "Must specify EITHER manifest_yaml (for YAML connectors) OR "
                    "docker_image + docker_tag (for Docker connectors), but not both"
                ),
                context={
                    "manifest_yaml_provided": is_yaml,
                    "docker_image_provided": is_docker,
                },
            )

        if is_docker and docker_tag is None:
            raise exc.PyAirbyteInputError(
                message="docker_tag is required when docker_image is specified",
                context={"docker_image": docker_image},
            )

        if unique:
            existing = self.list_custom_source_definitions(
                definition_type="yaml" if is_yaml else "docker",
            )
            if any(d.name == name for d in existing):
                raise exc.AirbyteDuplicateResourcesError(
                    resource_type="custom_source_definition",
                    resource_name=name,
                )

        if is_yaml:
            manifest_dict: dict[str, Any]
            if isinstance(manifest_yaml, Path):
                manifest_dict = yaml.safe_load(manifest_yaml.read_text())
            elif isinstance(manifest_yaml, str):
                manifest_dict = yaml.safe_load(manifest_yaml)
            elif manifest_yaml is not None:
                manifest_dict = manifest_yaml
            else:
                raise exc.PyAirbyteInputError(
                    message="manifest_yaml is required for YAML connectors",
                    context={"name": name},
                )

            if pre_validate:
                api_util.validate_yaml_manifest(manifest_dict, raise_on_error=True)

            result = api_util.create_custom_yaml_source_definition(
                name=name,
                workspace_id=self.workspace_id,
                manifest=manifest_dict,
                api_root=self.api_root,
                client_id=self.client_id,
                client_secret=self.client_secret,
                bearer_token=self.bearer_token,
            )
            custom_definition = CustomCloudSourceDefinition._from_yaml_response(  # noqa: SLF001
                self, result
            )

            # Set testing values if provided
            if testing_values is not None:
                custom_definition.set_testing_values(testing_values)

            return custom_definition

        raise NotImplementedError(
            "Docker custom source definitions are not yet supported. "
            "Only YAML manifest-based custom sources are currently available."
        )

    def list_custom_source_definitions(
        self,
        *,
        definition_type: Literal["yaml", "docker"],
    ) -> list[CustomCloudSourceDefinition]:
        """List custom source connector definitions.

        Args:
            definition_type: Connector type to list ("yaml" or "docker"). Required.

        Returns:
            List of CustomCloudSourceDefinition objects matching the specified type
        """
        if definition_type == "yaml":
            yaml_definitions = api_util.list_custom_yaml_source_definitions(
                workspace_id=self.workspace_id,
                api_root=self.api_root,
                client_id=self.client_id,
                client_secret=self.client_secret,
                bearer_token=self.bearer_token,
            )
            return [
                CustomCloudSourceDefinition._from_yaml_response(self, d)  # noqa: SLF001
                for d in yaml_definitions
            ]

        raise NotImplementedError(
            "Docker custom source definitions are not yet supported. "
            "Only YAML manifest-based custom sources are currently available."
        )

    def get_custom_source_definition(
        self,
        definition_id: str,
        *,
        definition_type: Literal["yaml", "docker"],
    ) -> CustomCloudSourceDefinition:
        """Get a specific custom source definition by ID.

        Args:
            definition_id: The definition ID
            definition_type: Connector type ("yaml" or "docker"). Required.

        Returns:
            CustomCloudSourceDefinition object
        """
        if definition_type == "yaml":
            result = api_util.get_custom_yaml_source_definition(
                workspace_id=self.workspace_id,
                definition_id=definition_id,
                api_root=self.api_root,
                client_id=self.client_id,
                client_secret=self.client_secret,
                bearer_token=self.bearer_token,
            )
            return CustomCloudSourceDefinition._from_yaml_response(self, result)  # noqa: SLF001

        raise NotImplementedError(
            "Docker custom source definitions are not yet supported. "
            "Only YAML manifest-based custom sources are currently available."
        )



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""Test a sync to an Airbyte destination.

Usage:
```
poetry run python examples/run_sync_to_destination_w_cache.py
```
"""

from __future__ import annotations

import datetime

import airbyte as ab

SCALE = 200_000


def get_my_source() -> ab.Source:
    return ab.get_source(
        "source-faker",
        config={
            "count": SCALE,
            "seed": 1234,
            "parallelism": 16,
        },
        streams="*",
    )


def get_my_destination() -> ab.Destination:
    return ab.get_destination(
        name="destination-duckdb",
        config={
            # This path is relative to the container:
            "destination_path": "/local/temp/db.duckdb",
        },
        docker_image=True,
        # OR:
        # pip_url="git+https://github.com/airbytehq/airbyte.git#subdirectory=airbyte-integrations/connectors/destination-duckdb",
    )


def main() -> None:
    """Test writing from the source to the destination."""
    source = get_my_source()
    source.check()
    destination = get_my_destination()
    destination.check()
    write_result: ab.WriteResult = destination.write(
        source_data=source,
        cache=ab.new_local_cache(),
    )
    print(
        f"Completed writing {write_result.processed_records:,} records "
        f"to destination at {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}."
    )


if __name__ == "__main__":
    main()



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.

"""Integration tests which leverage the source-faker connector to test the framework end-to-end.

Since source-faker is included in dev dependencies, we can assume `source-faker` is installed
and available on PATH for the poetry-managed venv.
"""

from __future__ import annotations

import os
import sys
from collections.abc import Generator
from pathlib import Path

import airbyte as ab
import pytest
from airbyte._util.venv_util import get_bin_dir
from airbyte.caches.duckdb import DuckDBCache
from airbyte.caches.util import new_local_cache


# Product count is always the same, regardless of faker scale.
NUM_PRODUCTS = 100

SEED_A = 1234
SEED_B = 5678

# Number of records in each of the 'users' and 'purchases' streams.
FAKER_SCALE_A = 200
# We want this to be different from FAKER_SCALE_A.
FAKER_SCALE_B = 300


@pytest.fixture(autouse=True)
def add_venv_bin_to_path(monkeypatch):
    """Patch the PATH to include the virtual environment's bin directory."""
    # Get the path to the bin directory of the virtual environment
    venv_bin_path = str(get_bin_dir(Path(sys.prefix)))

    # Add the bin directory to the PATH
    new_path = f"{venv_bin_path}{os.pathsep}{os.environ['PATH']}"
    monkeypatch.setenv("PATH", new_path)


def setup_source_faker(*, use_docker: bool) -> ab.Source:
    """Test the source-faker setup."""
    source = ab.get_source(
        "source-faker",
        config={
            "count": FAKER_SCALE_A,
            "seed": SEED_A,
            "parallelism": 16,  # Otherwise defaults to 4.
        },
        docker_image=use_docker,
    )
    source.check()
    source.select_streams([
        "users",
        "products",
        "purchases",
    ])
    return source


@pytest.fixture(scope="function")  # Each test gets a fresh source-faker instance.
def source_faker(*, use_docker: bool) -> ab.Source:
    """Fixture to return a source-faker connector instance."""
    return setup_source_faker(use_docker=use_docker)


def test_setup_source_faker(*, use_docker: bool) -> None:
    """Test that fixture logic works as expected."""
    source = setup_source_faker(use_docker=use_docker)


@pytest.fixture(scope="function")
def duckdb_cache() -> Generator[DuckDBCache, None, None]:
    """Fixture to return a fresh cache."""
    cache: DuckDBCache = new_local_cache()
    yield cache
    # TODO: Delete cache DB file after test is complete.
    return


def test_duckdb_connection_cleanup(tmp_path):
    """Test that DuckDB connections are properly released after cache operations.

    This test verifies the fix for issue #807: DuckDB Cache Does Not Release Database Connections.

    Reproduction steps:
    1. Create a DuckDB cache and write data to it
    2. Delete the cache object or call close()
    3. Attempt to open a new connection to the same database file
    4. Verify the new connection succeeds without errors
    """
    import duckdb

    db_path = tmp_path / "test_connection_cleanup.duckdb"

    cache = DuckDBCache(db_path=db_path)

    cache.processor._ensure_schema_exists()

    cache.processor._execute_sql(
        "CREATE TABLE IF NOT EXISTS test_table (id INTEGER, name VARCHAR)"
    )
    cache.processor._execute_sql("INSERT INTO test_table VALUES (1, 'test')")

    cache.close()

    try:
        conn = duckdb.connect(str(db_path))
        result = conn.execute("SELECT * FROM main.test_table").fetchall()
        assert len(result) == 1
        assert result[0] == (1, "test")
        conn.close()
    except Exception as e:
        pytest.fail(f"Failed to open new connection after cache cleanup: {e}")


def test_duckdb_context_manager_cleanup(tmp_path):
    """Test that DuckDB connections are cleaned up when using cache as context manager."""
    import duckdb

    db_path = tmp_path / "test_context_manager.duckdb"

    with DuckDBCache(db_path=db_path) as cache:
        cache.processor._execute_sql(
            "CREATE TABLE IF NOT EXISTS test_table (id INTEGER)"
        )
        cache.processor._execute_sql("INSERT INTO test_table VALUES (1)")

    conn = duckdb.connect(str(db_path))
    result = conn.execute("SELECT * FROM main.test_table").fetchall()
    assert len(result) == 1
    conn.close()


def test_duckdb_del_cleanup(tmp_path):
    """Test that DuckDB connections are cleaned up via __del__ when cache is garbage collected."""
    import duckdb
    import gc

    db_path = tmp_path / "test_del_cleanup.duckdb"

    def create_and_use_cache():
        cache = DuckDBCache(db_path=db_path)
        cache.processor._execute_sql(
            "CREATE TABLE IF NOT EXISTS test_table (id INTEGER)"
        )
        cache.processor._execute_sql("INSERT INTO test_table VALUES (1)")

    create_and_use_cache()

    gc.collect()

    conn = duckdb.connect(str(db_path))
    result = conn.execute("SELECT * FROM main.test_table").fetchall()
    assert len(result) == 1
    conn.close()



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Module which defines the `ReadResult` and `WriteResult` classes.

These classes are used to return information about read and write operations, respectively. They
contain information such as the number of records read or written, the cache object, and the
state handlers for a sync.
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import TYPE_CHECKING

from airbyte.datasets import CachedDataset


if TYPE_CHECKING:
    from collections.abc import Iterator

    from sqlalchemy.engine import Engine

    from airbyte._writers.base import AirbyteWriterInterface
    from airbyte.caches import CacheBase
    from airbyte.destinations.base import Destination
    from airbyte.progress import ProgressTracker
    from airbyte.shared.catalog_providers import CatalogProvider
    from airbyte.shared.state_providers import StateProviderBase
    from airbyte.shared.state_writers import StateWriterBase
    from airbyte.sources.base import Source


class ReadResult(Mapping[str, CachedDataset]):
    """The result of a read operation.

    This class is used to return information about the read operation, such as the number of
    records read. It should not be created directly, but instead returned by the write method
    of a destination.
    """

    def __init__(
        self,
        *,
        source_name: str,
        processed_streams: list[str],
        cache: CacheBase,
        progress_tracker: ProgressTracker,
    ) -> None:
        """Initialize a read result.

        This class should not be created directly. Instead, it should be returned by the `read`
        method of the `Source` class.
        """
        self.source_name = source_name
        self._progress_tracker = progress_tracker
        self._cache = cache
        self._processed_streams = processed_streams

    def __getitem__(self, stream: str) -> CachedDataset:
        """Return the cached dataset for a given stream name."""
        if stream not in self._processed_streams:
            raise KeyError(stream)

        return CachedDataset(self._cache, stream)

    def __contains__(self, stream: object) -> bool:
        """Return whether a given stream name was included in processing."""
        if not isinstance(stream, str):
            return False

        return stream in self._processed_streams

    def __iter__(self) -> Iterator[str]:
        """Return an iterator over the stream names that were processed."""
        return self._processed_streams.__iter__()

    def __len__(self) -> int:
        """Return the number of streams that were processed."""
        return len(self._processed_streams)

    def get_sql_engine(self) -> Engine:
        """Return the SQL engine used by the cache."""
        return self._cache.get_sql_engine()

    @property
    def processed_records(self) -> int:
        """The total number of records read from the source."""
        return self._progress_tracker.total_records_read

    @property
    def streams(self) -> Mapping[str, CachedDataset]:
        """Return a mapping of stream names to cached datasets."""
        return {
            stream_name: CachedDataset(self._cache, stream_name)
            for stream_name in self._processed_streams
        }

    @property
    def cache(self) -> CacheBase:
        """Return the cache object."""
        return self._cache


class WriteResult:
    """The result of a write operation.

    This class is used to return information about the write operation, such as the number of
    records written. It should not be created directly, but instead returned by the write method
    of a destination.
    """

    def __init__(
        self,
        *,
        destination: AirbyteWriterInterface | Destination,
        source_data: Source | ReadResult,
        catalog_provider: CatalogProvider,
        state_writer: StateWriterBase,
        progress_tracker: ProgressTracker,
    ) -> None:
        """Initialize a write result.

        This class should not be created directly. Instead, it should be returned by the `write`
        method of the `Destination` class.
        """
        self._destination: AirbyteWriterInterface | Destination = destination
        self._source_data: Source | ReadResult = source_data
        self._catalog_provider: CatalogProvider = catalog_provider
        self._state_writer: StateWriterBase = state_writer
        self._progress_tracker: ProgressTracker = progress_tracker

    @property
    def processed_records(self) -> int:
        """The total number of records written to the destination."""
        return self._progress_tracker.total_destination_records_delivered

    def get_state_provider(self) -> StateProviderBase:
        """Return the state writer as a state provider.

        As a public interface, we only expose the state writer as a state provider. This is because
        the state writer itself is only intended for internal use. As a state provider, the state
        writer can be used to read the state artifacts that were written. This can be useful for
        testing or debugging.
        """
        return self._state_writer



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
from __future__ import annotations

from itertools import islice

import airbyte as ab


# preparation (from PyAirbyte main folder):
#   python -m venv .venv-source-spacex-api
#   source .venv-source-spacex-api/bin/activate
#   pip install -e ../airbyte-integrations/connectors/source-spacex-api
# In separate terminal:
#   poetry run python examples/run_spacex.py

source = ab.get_source(
    "source-spacex-api",
    config={"id": "605b4b6aaa5433645e37d03f"},
    install_if_missing=True,
)
cache = ab.new_local_cache()

source.check()

source.select_streams(["launches", "rockets", "capsules"])

result = source.read(cache)

print(islice(source.get_records("capsules"), 10))

for name, records in result.cache.streams.items():
    print(f"Stream {name}: {len(list(records))} records")



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Unit tests for MCP tool utility functions."""

from __future__ import annotations

from unittest.mock import patch

import pytest

from airbyte.mcp._tool_utils import (
    SafeModeError,
    _GUIDS_CREATED_IN_SESSION,
    check_guid_created_in_session,
    register_guid_created_in_session,
)


@pytest.fixture(autouse=True)
def clear_session_guids() -> None:
    """Clear the session GUIDs before each test."""
    _GUIDS_CREATED_IN_SESSION.clear()


def test_register_guid_created_in_session() -> None:
    """Test that GUIDs can be registered as created in session."""
    assert "test-guid-123" not in _GUIDS_CREATED_IN_SESSION
    register_guid_created_in_session("test-guid-123")
    assert "test-guid-123" in _GUIDS_CREATED_IN_SESSION


def test_check_guid_created_in_session_passes_for_registered_guid() -> None:
    """Test that check passes for GUIDs registered in session."""
    register_guid_created_in_session("test-guid-456")
    # Should not raise
    check_guid_created_in_session("test-guid-456")


def test_check_guid_created_in_session_raises_for_unregistered_guid() -> None:
    """Test that check raises SafeModeError for unregistered GUIDs when safe mode is enabled."""
    with patch("airbyte.mcp._tool_utils.AIRBYTE_CLOUD_MCP_SAFE_MODE", True):
        with pytest.raises(SafeModeError) as exc_info:
            check_guid_created_in_session("unregistered-guid")
        assert "unregistered-guid" in str(exc_info.value)
        assert "not created in this session" in str(exc_info.value)


def test_check_guid_created_in_session_passes_when_safe_mode_disabled() -> None:
    """Test that check passes for any GUID when safe mode is disabled."""
    with patch("airbyte.mcp._tool_utils.AIRBYTE_CLOUD_MCP_SAFE_MODE", False):
        # Should not raise even for unregistered GUID
        check_guid_created_in_session("any-guid-at-all")


def test_multiple_guids_can_be_registered() -> None:
    """Test that multiple GUIDs can be registered in the same session."""
    guids = ["guid-1", "guid-2", "guid-3"]
    for guid in guids:
        register_guid_created_in_session(guid)

    for guid in guids:
        assert guid in _GUIDS_CREATED_IN_SESSION


def test_duplicate_guid_registration_is_idempotent() -> None:
    """Test that registering the same GUID multiple times is safe."""
    register_guid_created_in_session("duplicate-guid")
    register_guid_created_in_session("duplicate-guid")
    assert "duplicate-guid" in _GUIDS_CREATED_IN_SESSION



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""NoOp executor for connectors that don't require local execution."""

from __future__ import annotations

import json
import logging
from typing import TYPE_CHECKING, Any

from airbyte import exceptions as exc
from airbyte._executors.base import Executor
from airbyte._util.registry_spec import get_connector_spec_from_registry


if TYPE_CHECKING:
    from collections.abc import Iterator
    from typing import IO

    from airbyte._message_iterators import AirbyteMessageIterator
    from airbyte.registry import ConnectorMetadata


logger = logging.getLogger("airbyte")


class NoOpExecutor(Executor):
    """Executor that fetches specs from registry without local installation.

    This executor is useful for scenarios where you need to validate connector
    configurations but don't need to actually run the connector locally (e.g.,
    when deploying to Airbyte Cloud).

    The NoOpExecutor:
    - Fetches connector specs from the Airbyte registry
    - Supports the 'spec' command for configuration validation
    - Does not support execution commands (check, discover, read, write)
    - Does not require Docker or Python installation
    """

    def __init__(
        self,
        *,
        name: str | None = None,
        metadata: ConnectorMetadata | None = None,
        target_version: str | None = None,
    ) -> None:
        """Initialize the NoOp executor."""
        super().__init__(
            name=name,
            metadata=metadata,
            target_version=target_version,
        )
        self._cached_spec: dict[str, Any] | None = None

    @property
    def _cli(self) -> list[str]:
        """Return a placeholder CLI command.

        This is never actually used since execute() is overridden.
        """
        return ["noop"]

    def execute(
        self,
        args: list[str],
        *,
        stdin: IO[str] | AirbyteMessageIterator | None = None,
        suppress_stderr: bool = False,
    ) -> Iterator[str]:
        """Execute a command and return an iterator of STDOUT lines.

        Only the 'spec' command is supported. Other commands will raise an error.
        """
        _ = stdin, suppress_stderr  # Unused

        if args == ["spec"]:
            if self._cached_spec is None:
                spec = get_connector_spec_from_registry(
                    self.name,
                    version=self.target_version,
                    platform="cloud",
                )
                if spec is None:
                    spec = get_connector_spec_from_registry(
                        self.name,
                        version=self.target_version,
                        platform="oss",
                    )

                if spec is None:
                    logger.warning(
                        f"Could not fetch spec for connector '{self.name}' from registry. "
                        f"Pre-validation will not be available."
                    )
                    yield from []
                    return

                self._cached_spec = spec

            spec_message = {
                "type": "SPEC",
                "spec": {
                    "connectionSpecification": self._cached_spec,
                },
            }
            yield json.dumps(spec_message)
            return

        raise exc.AirbyteConnectorExecutableNotFoundError(
            connector_name=self.name,
            guidance=(
                "NoOpExecutor only supports the 'spec' command for configuration validation. "
                "To run connector operations (check, discover, read), "
                "use a Docker or Python executor. "
                "For MCP cloud deployments, don't use no_executor=True "
                "if you need to run connector operations locally."
            ),
        )

    def ensure_installation(self, *, auto_fix: bool = True) -> None:
        """No-op: NoOpExecutor doesn't require installation."""
        _ = auto_fix  # Unused
        pass

    def install(self) -> None:
        """No-op: NoOpExecutor doesn't require installation."""
        pass

    def uninstall(self) -> None:
        """No-op: NoOpExecutor doesn't manage installations."""
        pass


__all__ = [
    "NoOpExecutor",
]



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Unit tests for MCP connector registry tools."""

from __future__ import annotations

from unittest.mock import MagicMock, patch

from airbyte import exceptions as exc
from airbyte.mcp.registry import get_api_docs_urls
from airbyte.registry import (
    ApiDocsUrl,
    _fetch_manifest_dict,
    _manifest_url_for,
)


class TestManifestUrlFor:
    """Tests for _manifest_url_for function."""

    def test_manifest_url_for(self) -> None:
        """Test generating manifest URL for a connector."""
        url = _manifest_url_for("source-example")
        assert "source-example" in url
        assert "manifest.yaml" in url
        assert "latest" in url


class TestFetchManifestDict:
    """Tests for _fetch_manifest_dict function."""

    def test_manifest_not_found(self) -> None:
        """Test handling when manifest.yaml doesn't exist (404)."""
        with patch("airbyte.registry.requests.get") as mock_get:
            mock_response = MagicMock()
            mock_response.status_code = 404
            mock_get.return_value = mock_response

            manifest_dict = _fetch_manifest_dict("https://example.com/manifest.yaml")
            assert manifest_dict == {}

    def test_fetch_manifest_dict(self) -> None:
        """Test fetching and parsing manifest.yaml."""
        manifest_yaml = """
version: 1.0.0
type: DeclarativeSource
data:
  name: Example
"""
        with patch("airbyte.registry.requests.get") as mock_get:
            mock_response = MagicMock()
            mock_response.status_code = 200
            mock_response.text = manifest_yaml
            mock_get.return_value = mock_response

            manifest_dict = _fetch_manifest_dict("https://example.com/manifest.yaml")
            assert manifest_dict["version"] == "1.0.0"
            assert manifest_dict["type"] == "DeclarativeSource"
            assert manifest_dict["data"]["name"] == "Example"


class TestApiDocsUrlFromManifestDict:
    """Tests for ApiDocsUrl.from_manifest_dict classmethod."""

    def test_manifest_with_external_docs_urls(self) -> None:
        """Test extracting URLs from data.externalDocumentationUrls field."""
        manifest_dict = {
            "version": "1.0.0",
            "type": "DeclarativeSource",
            "data": {
                "externalDocumentationUrls": [
                    {
                        "title": "Versioning docs",
                        "url": "https://api.example.com/versioning",
                        "type": "api_reference",
                    },
                    {
                        "title": "Changelog",
                        "url": "https://api.example.com/changelog",
                        "type": "api_release_history",
                    },
                    {
                        "title": "Deprecated API calls",
                        "url": "https://api.example.com/deprecations",
                        "type": "api_deprecations",
                        "requiresLogin": True,
                    },
                ]
            },
        }

        urls = ApiDocsUrl.from_manifest_dict(manifest_dict)
        assert len(urls) == 3
        assert urls[0].title == "Versioning docs"
        assert urls[0].url == "https://api.example.com/versioning"
        assert urls[0].doc_type == "api_reference"
        assert urls[0].requires_login is False
        assert urls[1].title == "Changelog"
        assert urls[1].doc_type == "api_release_history"
        assert urls[2].title == "Deprecated API calls"
        assert urls[2].doc_type == "api_deprecations"
        assert urls[2].requires_login is True

    def test_manifest_with_external_docs_no_type(self) -> None:
        """Test extracting URLs from data.externalDocumentationUrls without type field."""
        manifest_dict = {
            "version": "1.0.0",
            "type": "DeclarativeSource",
            "data": {
                "externalDocumentationUrls": [
                    {
                        "title": "General docs",
                        "url": "https://api.example.com/docs",
                    }
                ]
            },
        }

        urls = ApiDocsUrl.from_manifest_dict(manifest_dict)
        assert len(urls) == 1
        assert urls[0].title == "General docs"
        assert urls[0].doc_type == "other"
        assert urls[0].requires_login is False

    def test_empty_manifest(self) -> None:
        """Test handling empty manifest dict."""
        urls = ApiDocsUrl.from_manifest_dict({})
        assert len(urls) == 0


class TestGetApiDocsUrls:
    """Tests for get_api_docs_urls function."""

    def test_connector_not_found(self) -> None:
        """Test handling when connector is not found."""
        with patch("airbyte.mcp.registry.get_connector_api_docs_urls") as mock_get_docs:
            mock_get_docs.side_effect = exc.AirbyteConnectorNotRegisteredError(
                connector_name="nonexistent-connector",
                context={},
            )

            result = get_api_docs_urls("nonexistent-connector")
            assert result == "Connector not found."

    def test_deduplication_of_urls(self) -> None:
        """Test that duplicate URLs are deduplicated."""
        with patch("airbyte.mcp.registry.get_connector_api_docs_urls") as mock_get_docs:
            mock_get_docs.return_value = [
                ApiDocsUrl(
                    title="Airbyte Documentation",
                    url="https://docs.airbyte.com/integrations/sources/example",
                    source="registry",
                )
            ]

            result = get_api_docs_urls("source-example")

            assert isinstance(result, list)
            assert len(result) == 1
            assert result[0].title == "Airbyte Documentation"



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""Global pytest fixtures."""

from __future__ import annotations

import logging
import os
import shutil
import socket
import sys
import time
import warnings
from pathlib import Path

import airbyte
import docker
import psycopg
import pytest
from _pytest.nodes import Item
from airbyte._executors.util import get_connector_executor
from airbyte._util import text_util
from airbyte._util.meta import is_docker_installed, is_windows
from airbyte.caches import PostgresCache
from airbyte.caches.duckdb import DuckDBCache
from airbyte.caches.util import new_local_cache
from requests.exceptions import HTTPError


logger = logging.getLogger(__name__)


PYTEST_POSTGRES_IMAGE = "postgres:13"
PYTEST_POSTGRES_CONTAINER = "postgres_pytest_container"
PYTEST_POSTGRES_PORT = 5432

LOCAL_TEST_REGISTRY_URL = "./tests/integration_tests/fixtures/registry.json"
USE_DOCKER = is_docker_installed() and not is_windows()


@pytest.fixture
def use_docker() -> bool:
    return USE_DOCKER


def pytest_configure(config):
    if sys.platform.startswith("win"):
        # Ignore specific warnings only on Windows
        warnings.filterwarnings(
            "ignore",
            category=pytest.PytestUnraisableExceptionWarning,
        )
        warnings.filterwarnings(
            "ignore",
            category=ResourceWarning,
            message="unclosed file",
        )


def pytest_collection_modifyitems(items: list[Item]) -> None:
    """Override default pytest behavior, sorting our tests in a sensible execution order.

    In general, we want faster tests to run first, so that we can get feedback faster.

    Running lint tests first is helpful because they are fast and can catch typos and other errors.

    Otherwise tests are run based on an alpha-based natural sort, where 'unit' tests run after
    'integration' tests because 'u' comes after 'i' alphabetically.
    """

    def test_priority(item: Item) -> int:
        if item.get_closest_marker(name="slow"):
            return 9  # slow tests have the lowest priority
        elif "lint_tests" in str(item.fspath):
            return 1  # lint tests have high priority
        elif "unit_tests" in str(item.fspath):
            return 2  # unit tests have highest priority
        elif "docs_tests" in str(item.fspath):
            return 3  # doc tests have medium priority
        elif "integration_tests" in str(item.fspath):
            return 4  # integration tests have the lowest priority
        else:
            return 5  # all other tests have lower priority

    # Sort the items list in-place based on the test_priority function
    items.sort(key=test_priority)

    for item in items:
        # Skip tests that require Docker if Docker is not available (including on Windows).
        if (
            "new_postgres_cache" in item.fixturenames
            or "postgres_cache" in item.fixturenames
            or "source_docker_faker_seed_a" in item.fixturenames
            or "source_docker_faker_seed_b" in item.fixturenames
            or "new_duckdb_destination_executor" in item.fixturenames
            or "e2e_test_destination" in item.fixturenames
        ):
            if not is_docker_available():
                item.add_marker(
                    pytest.mark.skip(reason="Skipping tests (Docker not available)")
                )

        # Every test in the cloud directory is slow abd requires credentials
        if "integration_tests/cloud" in str(item.fspath):
            item.add_marker(pytest.mark.slow)
            item.add_marker(pytest.mark.requires_creds)

        if "super_slow" in item.keywords:
            # Super slow tests are also slow
            item.add_marker("slow")


def is_port_in_use(port):
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        return s.connect_ex(("localhost", port)) == 0


@pytest.fixture(scope="session", autouse=True)
def remove_postgres_container():
    if is_port_in_use(PYTEST_POSTGRES_PORT):
        try:
            client = docker.from_env()
            container = client.containers.get(
                PYTEST_POSTGRES_CONTAINER,
            )
            container.stop()
            container.remove()
        except docker.errors.NotFound:
            pass  # Container not found, nothing to do.
        except docker.errors.DockerException:
            pass  # Docker not running, nothing to do.


def test_pg_connection(host) -> bool:
    pg_url = f"postgresql://postgres:postgres@{host}:{PYTEST_POSTGRES_PORT}/postgres"

    max_attempts = 120
    for attempt in range(max_attempts):
        try:
            conn = psycopg.connect(pg_url)
            conn.close()
            return True
        except psycopg.OperationalError:
            logger.info(
                f"Waiting for postgres to start (attempt {attempt + 1}/{max_attempts})"
            )
            time.sleep(1.0)

    else:
        return False


def is_docker_available():
    if is_windows():
        # Linux containers are not supported on Windows CI runners
        return False
    try:
        _ = docker.from_env()
        return True
    except docker.errors.DockerException:
        return False


@pytest.fixture(scope="session")
def new_postgres_db():
    """Fixture to start a new PostgreSQL container for testing.

    This fixture will start a new PostgreSQL container before the tests run and stop it after the
    tests are done. The host of the PostgreSQL database will be returned to the tests.
    """
    client = docker.from_env()
    try:
        client.images.get(PYTEST_POSTGRES_IMAGE)
    except (docker.errors.ImageNotFound, HTTPError):
        # Pull the image if it doesn't exist, to avoid failing our sleep timer
        # if the image needs to download on-demand.
        client.images.pull(PYTEST_POSTGRES_IMAGE)

    try:
        previous_container = client.containers.get(PYTEST_POSTGRES_CONTAINER)
        previous_container.remove()
    except docker.errors.NotFound:
        pass

    postgres_is_running = False
    postgres = client.containers.run(
        image=PYTEST_POSTGRES_IMAGE,
        name=PYTEST_POSTGRES_CONTAINER,
        environment={
            "POSTGRES_USER": "postgres",
            "POSTGRES_PASSWORD": "postgres",
            "POSTGRES_DB": "postgres",
        },
        ports={"5432/tcp": PYTEST_POSTGRES_PORT},
        detach=True,
    )

    attempts = 10
    while not postgres_is_running and attempts > 0:
        try:
            postgres.reload()
            postgres_is_running = postgres.status == "running"
        except docker.errors.NotFound:
            attempts -= 1
            time.sleep(3)
    if not postgres_is_running:
        raise Exception(
            f"Failed to start the PostgreSQL container. Status: {postgres.status}."
        )

    final_host = None
    if host := os.environ.get("DOCKER_HOST_NAME"):
        final_host = host if test_pg_connection(host) else None
    else:
        # Try to connect to the database using localhost and the docker host IP
        for host in ["127.0.0.1", "localhost", "host.docker.internal", "172.17.0.1"]:
            if test_pg_connection(host):
                final_host = host
                break

    if final_host is None:
        raise Exception(f"Failed to connect to the PostgreSQL database on host {host}.")

    yield final_host

    # Stop and remove the container after the tests are done
    postgres.stop()
    postgres.remove()


@pytest.fixture(scope="function")
def new_postgres_cache(new_postgres_db: str):
    """Fixture to return a fresh Postgres cache.

    Each test that uses this fixture will get a unique table prefix.
    """
    config = PostgresCache(
        host=new_postgres_db,
        port=PYTEST_POSTGRES_PORT,
        username="postgres",
        password="postgres",
        database="postgres",
        schema_name="public",
        # TODO: Move this to schema name when we support it (breaks as of 2024-01-31):
        table_prefix=f"test{text_util.generate_random_suffix()}_",
    )
    yield config


@pytest.fixture(autouse=False)
def source_test_registry(monkeypatch):
    """Mock the registry to return our custom registry containing the 'source-test' connector.

    This means the normal registry is not usable. Expect AirbyteConnectorNotRegisteredError for
    other connectors.
    """

    # Define the mock function
    def mock_get_registry_cache():
        return LOCAL_TEST_REGISTRY_URL

    # Replace _get_registry_url() with the mock function
    monkeypatch.setattr(airbyte.registry, "_get_registry_url", mock_get_registry_cache)

    # reset the registry cache
    airbyte.registry.__cache = None

    yield

    # reset the registry cache (clean up)
    airbyte.registry.__cache = None


@pytest.fixture(autouse=True)
def do_not_track(monkeypatch):
    """Set environment variables for the test source.

    These are applied to this test file only.
    """
    env_vars = {"DO_NOT_TRACK": "true"}
    for key, value in env_vars.items():
        monkeypatch.setenv(key, value)


@pytest.fixture(
    scope="session", params=[True, False], ids=["uv_enabled", "uv_disabled"]
)
def source_test_installation(request):
    """Test fixture for sample source installation.

    Prepare test environment. This will pre-install the test source from the fixtures array and set
    the environment variable to use the local json file as registry.

    Parametrized to test both uv-enabled and uv-disabled installation methods.
    """
    use_uv = request.param

    if not use_uv:
        os.environ["AIRBYTE_NO_UV"] = "1"

    venv_dir = ".venv-source-test"
    if Path(venv_dir).exists():
        shutil.rmtree(venv_dir)

    try:
        executor = get_connector_executor(
            name="source-test",
            pip_url="./tests/integration_tests/fixtures/source-test",
            install_root=Path.cwd(),
            install_if_missing=True,
        )

        yield executor

    finally:
        if Path(venv_dir).exists():
            shutil.rmtree(venv_dir)
        if not use_uv and "AIRBYTE_NO_UV" in os.environ:
            del os.environ["AIRBYTE_NO_UV"]


@pytest.fixture(scope="function")
def new_duckdb_cache() -> DuckDBCache:
    return new_local_cache()



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
"""A sample execution script which loads data from `source-faker` to a MotherDuck-backed cache.

Usage (from repo root):
    poetry install
    poetry run python examples/run_faker_to_motherduck.py
"""

from __future__ import annotations

import airbyte as ab
from airbyte.caches import MotherDuckCache


MOTHERDUCK_API_KEY = ab.get_secret("MOTHERDUCK_API_KEY")
"""This is the API key for the MotherDuck service.

It can be auto-detected in env vars and/or a .env file in the root of the project.

If will be prompted (and masked during input) if not found in either location.
"""


source = ab.get_source(
    "source-faker",
    config={"count": 10000, "seed": 0, "parallelism": 1, "always_updated": False},
    install_if_missing=True,
)
source.check()
source.select_all_streams()

cache = MotherDuckCache(
    database="airbyte_test",
    schema_name="faker_data",
    api_key=MOTHERDUCK_API_KEY,
)

result = source.read(cache=cache, force_full_refresh=True)

for name, records in result.streams.items():
    print(f"Stream {name}: {len(records)} records")



"""Simple script to download secrets from GCS.

Secrets will be located based on the `connector` label in the GSM secret metadata, and they
will be written to the connector's secrets directory based upon the `filename` label.

Filename is appended with `.json` and the secret is written to that file.

As a safety measure, we will only write to the connector's secrets directory if it exists.
If it doesn't exist, the script will fail. Users should ensure the directory
exists and is excluded from git before running the script.

Usage:
    poetry run python examples/run_gsm_connector_secret_fetch.py
    poetry run python examples/run_gsm_connector_secret_fetch.py <connector_name>
    poetry run python examples/run_gsm_connector_secret_fetch.py source-github
"""

from __future__ import annotations

import sys
from pathlib import Path

import airbyte as ab
from airbyte.secrets import GoogleGSMSecretManager, SecretHandle

AIRBYTE_INTERNAL_GCP_PROJECT = "dataline-integration-testing"
CONNECTOR_NAME = sys.argv[1] if len(sys.argv) > 1 else "source-klaviyo"

AIRBYTE_REPO_ROOT = Path(__file__).parent.parent.parent / "airbyte"


CONNECTOR_SECRETS_DIR = (
    AIRBYTE_REPO_ROOT
    / "airbyte-integrations"
    / "connectors"
    / CONNECTOR_NAME
    / "secrets"
)
if not AIRBYTE_REPO_ROOT.exists():
    raise FileNotFoundError(f"Airbyte repo root does not exist: {AIRBYTE_REPO_ROOT}")
if not CONNECTOR_SECRETS_DIR.exists():
    CONNECTOR_SECRETS_DIR.mkdir(parents=True, exist_ok=True)


def main() -> None:
    secret_mgr = GoogleGSMSecretManager(
        project=AIRBYTE_INTERNAL_GCP_PROJECT,
        credentials_json=ab.get_secret("GCP_GSM_CREDENTIALS"),
    )

    secret: SecretHandle
    for secret in secret_mgr.fetch_connector_secrets(
        connector_name=CONNECTOR_NAME,
    ):
        filename_base = "config"  # Default filename if not overridden
        if "filename" in secret.labels:
            filename_base = secret.labels["filename"]
        secret_file_path = CONNECTOR_SECRETS_DIR / f"{filename_base}.json"
        secret.write_to_file(secret_file_path)


if __name__ == "__main__":
    main()



# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
from __future__ import annotations

import pytest
from airbyte._util import github_pip_url, connector_pip_url


@pytest.mark.parametrize(
    "owner, repo, branch_or_ref, package_name, subdirectory, expected",
    [
        (
            "airbytehq",
            "airbyte",
            None,
            None,
            None,
            "git+https://github.com/airbytehq/airbyte.git",
        ),
        (
            "airbytehq",
            "airbyte",
            "master",
            None,
            None,
            "git+https://github.com/airbytehq/airbyte.git@master",
        ),
        (
            "airbytehq",
            "airbyte",
            "my-branch",
            None,
            None,
            "git+https://github.com/airbytehq/airbyte.git@my-branch",
        ),
        (
            "airbytehq",
            "airbyte",
            "my-branch",
            "airbyte-lib",
            None,
            "git+https://github.com/airbytehq/airbyte.git@my-branch#egg=airbyte-lib",
        ),
        (
            "airbytehq",
            "airbyte",
            "my-branch",
            "airbyte-lib",
            "airbyte-lib",
            "git+https://github.com/airbytehq/airbyte.git@my-branch#egg=airbyte-lib&subdirectory=airbyte-lib",
        ),
    ],
)
def test_github_pip_url(
    owner, repo, branch_or_ref, package_name, subdirectory, expected
):
    result = github_pip_url(
        owner,
        repo,
        branch_or_ref=branch_or_ref,
        package_name=package_name,
        subdirectory=subdirectory,
    )
    assert result == expected


@pytest.mark.parametrize(
    "connector_name, branch, owner, expected",
    [
        (
            "source-coin-api",
            "my-branch",
            None,
            "git+https://github.com/airbytehq/airbyte.git@my-branch#egg=source-coin-api&subdirectory=airbyte-integrations/connectors/source-coin-api",
        ),
        (
            "source-coin-api",
            "my-branch",
            "my-fork",
            "git+https://github.com/my-fork/airbyte.git@my-branch#egg=source-coin-api&subdirectory=airbyte-integrations/connectors/source-coin-api",
        ),
    ],
)
def test_connector_pip_url(connector_name, branch, owner, expected):
    result = connector_pip_url(connector_name, branch, owner=owner)
    assert result == expected



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.
"""Local MCP operations."""

import sys
import traceback
from itertools import islice
from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Any, Literal

from fastmcp import FastMCP
from fastmcp_extensions import mcp_tool, register_mcp_tools
from pydantic import BaseModel, Field

from airbyte import get_source
from airbyte._util.meta import is_docker_installed
from airbyte.caches.util import get_default_cache
from airbyte.mcp._arg_resolvers import resolve_connector_config, resolve_list_of_strings
from airbyte.registry import get_connector_metadata
from airbyte.secrets.config import _get_secret_sources
from airbyte.secrets.env_vars import DotenvSecretManager
from airbyte.secrets.google_gsm import GoogleGSMSecretManager
from airbyte.sources.base import Source


if TYPE_CHECKING:
    from airbyte.caches.duckdb import DuckDBCache


_CONFIG_HELP = """
You can provide `config` as JSON or a Path to a YAML/JSON file.
If a `dict` is provided, it must not contain hardcoded secrets.
Instead, secrets should be provided using environment variables,
and the config should reference them using the format
`secret_reference::ENV_VAR_NAME`.

You can also provide a `config_secret_name` to use a specific
secret name for the configuration. This is useful if you want to
validate a configuration that is stored in a secrets manager.

If `config_secret_name` is provided, it should point to a string
that contains valid JSON or YAML.

If both `config` and `config_secret_name` are provided, the
`config` will be loaded first and then the referenced secret config
will be layered on top of the non-secret config.

For declarative connectors, you can provide a `manifest_path` to
specify a local YAML manifest file instead of using the registry
version. This is useful for testing custom or locally-developed
connector manifests.
"""


def _get_mcp_source(
    connector_name: str,
    override_execution_mode: Literal["auto", "docker", "python", "yaml"] = "auto",
    *,
    install_if_missing: bool = True,
    manifest_path: str | Path | None,
) -> Source:
    """Get the MCP source for a connector."""
    if manifest_path:
        override_execution_mode = "yaml"
    elif override_execution_mode == "auto" and is_docker_installed():
        override_execution_mode = "docker"

    source: Source
    if override_execution_mode == "auto":
        # Use defaults with no overrides
        source = get_source(
            connector_name,
            install_if_missing=False,
            source_manifest=manifest_path or None,
        )
    elif override_execution_mode == "python":
        source = get_source(
            connector_name,
            use_python=True,
            install_if_missing=False,
            source_manifest=manifest_path or None,
        )
    elif override_execution_mode == "docker":
        source = get_source(
            connector_name,
            docker_image=True,
            install_if_missing=False,
            source_manifest=manifest_path or None,
        )
    elif override_execution_mode == "yaml":
        source = get_source(
            connector_name,
            source_manifest=manifest_path or True,
            install_if_missing=False,
        )
    else:
        raise ValueError(
            f"Unknown execution method: {override_execution_mode}. "
            "Expected one of: ['auto', 'docker', 'python', 'yaml']."
        )

    # Ensure installed:
    if install_if_missing:
        source.executor.ensure_installation()

    return source


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def validate_connector_config(
    connector_name: Annotated[
        str,
        Field(description="The name of the connector to validate."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the connector as a dict object or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the connector configuration.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> tuple[bool, str]:
    """Validate a connector configuration.

    Returns a tuple of (is_valid: bool, message: str).
    """
    try:
        source: Source = _get_mcp_source(
            connector_name,
            override_execution_mode=override_execution_mode,
            manifest_path=manifest_path,
        )
    except Exception as ex:
        return False, f"Failed to get connector '{connector_name}': {ex}"

    try:
        config_dict = resolve_connector_config(
            config=config,
            config_file=config_file,
            config_secret_name=config_secret_name,
            config_spec_jsonschema=source.config_spec,
        )
        source.set_config(config_dict)
    except Exception as ex:
        return False, f"Failed to resolve configuration for {connector_name}: {ex}"

    try:
        source.check()
    except Exception as ex:
        return False, f"Configuration for {connector_name} is invalid: {ex}"

    return True, f"Configuration for {connector_name} is valid!"


@mcp_tool(
    read_only=True,
    idempotent=True,
)
def list_connector_config_secrets(
    connector_name: Annotated[
        str,
        Field(description="The name of the connector."),
    ],
) -> list[str]:
    """List all `config_secret_name` options that are known for the given connector.

    This can be used to find out which already-created config secret names are available
    for a given connector. The return value is a list of secret names, but it will not
    return the actual secret values.
    """
    secrets_names: list[str] = []
    for secrets_mgr in _get_secret_sources():
        if isinstance(secrets_mgr, GoogleGSMSecretManager):
            secrets_names.extend(
                [
                    secret_handle.secret_name.split("/")[-1]
                    for secret_handle in secrets_mgr.fetch_connector_secrets(connector_name)
                ]
            )

    return secrets_names


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def list_dotenv_secrets() -> dict[str, list[str]]:
    """List all environment variable names declared within declared .env files.

    This returns a dictionary mapping the .env file name to a list of environment
    variable names. The values of the environment variables are not returned.
    """
    result: dict[str, list[str]] = {}
    for secrets_mgr in _get_secret_sources():
        if isinstance(secrets_mgr, DotenvSecretManager) and secrets_mgr.dotenv_path:
            result[str(secrets_mgr.dotenv_path.resolve())] = secrets_mgr.list_secrets_names()

    return result


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def list_source_streams(
    source_connector_name: Annotated[
        str,
        Field(description="The name of the source connector."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the source connector as a dict or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the source connector config.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> list[str]:
    """List all streams available in a source connector.

    This operation (generally) requires a valid configuration, including any required secrets.
    """
    source: Source = _get_mcp_source(
        connector_name=source_connector_name,
        override_execution_mode=override_execution_mode,
        manifest_path=manifest_path,
    )
    config_dict = resolve_connector_config(
        config=config,
        config_file=config_file,
        config_secret_name=config_secret_name,
        config_spec_jsonschema=source.config_spec,
    )
    source.set_config(config_dict)
    return source.get_available_streams()


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def get_source_stream_json_schema(
    source_connector_name: Annotated[
        str,
        Field(description="The name of the source connector."),
    ],
    stream_name: Annotated[
        str,
        Field(description="The name of the stream."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the source connector as a dict or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the source connector config.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> dict[str, Any]:
    """List all properties for a specific stream in a source connector."""
    source: Source = _get_mcp_source(
        connector_name=source_connector_name,
        override_execution_mode=override_execution_mode,
        manifest_path=manifest_path,
    )
    config_dict = resolve_connector_config(
        config=config,
        config_file=config_file,
        config_secret_name=config_secret_name,
        config_spec_jsonschema=source.config_spec,
    )
    source.set_config(config_dict)
    return source.get_stream_json_schema(stream_name=stream_name)


@mcp_tool(
    read_only=True,
    extra_help_text=_CONFIG_HELP,
)
def read_source_stream_records(
    source_connector_name: Annotated[
        str,
        Field(description="The name of the source connector."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the source connector as a dict or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the source connector config.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    *,
    stream_name: Annotated[
        str,
        Field(description="The name of the stream to read records from."),
    ],
    max_records: Annotated[
        int,
        Field(
            description="The maximum number of records to read.",
            default=1000,
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> list[dict[str, Any]] | str:
    """Get records from a source connector."""
    try:
        source: Source = _get_mcp_source(
            connector_name=source_connector_name,
            override_execution_mode=override_execution_mode,
            manifest_path=manifest_path,
        )
        config_dict = resolve_connector_config(
            config=config,
            config_file=config_file,
            config_secret_name=config_secret_name,
            config_spec_jsonschema=source.config_spec,
        )
        source.set_config(config_dict)
        # First we get a generator for the records in the specified stream.
        record_generator = source.get_records(stream_name)
        # Next we load a limited number of records from the generator into our list.
        records: list[dict[str, Any]] = list(islice(record_generator, max_records))

        print(f"Retrieved {len(records)} records from stream '{stream_name}'", sys.stderr)

    except Exception as ex:
        tb_str = traceback.format_exc()
        # If any error occurs, we print the error message to stderr and return an empty list.
        return (
            f"Error reading records from source '{source_connector_name}': {ex!r}, {ex!s}\n{tb_str}"
        )

    else:
        return records


@mcp_tool(
    read_only=True,
    extra_help_text=_CONFIG_HELP,
)
def get_stream_previews(
    source_name: Annotated[
        str,
        Field(description="The name of the source connector."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the source connector as a dict or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the source connector config.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    streams: Annotated[
        list[str] | str | None,
        Field(
            description=(
                "The streams to get previews for. "
                "Use '*' for all streams, or None for selected streams."
            ),
            default=None,
        ),
    ],
    limit: Annotated[
        int,
        Field(
            description="The maximum number of sample records to return per stream.",
            default=10,
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> dict[str, list[dict[str, Any]] | str]:
    """Get sample records (previews) from streams in a source connector.

    This operation requires a valid configuration, including any required secrets.
    Returns a dictionary mapping stream names to lists of sample records, or an error
    message string if an error occurred for that stream.
    """
    source: Source = _get_mcp_source(
        connector_name=source_name,
        override_execution_mode=override_execution_mode,
        manifest_path=manifest_path,
    )

    config_dict = resolve_connector_config(
        config=config,
        config_file=config_file,
        config_secret_name=config_secret_name,
        config_spec_jsonschema=source.config_spec,
    )
    source.set_config(config_dict)

    streams_param: list[str] | Literal["*"] | None = resolve_list_of_strings(
        streams
    )  # pyrefly: ignore[no-matching-overload]
    if streams_param and len(streams_param) == 1 and streams_param[0] == "*":
        streams_param = "*"

    try:
        samples_result = source.get_samples(
            streams=streams_param,
            limit=limit,
            on_error="ignore",
        )
    except Exception as ex:
        tb_str = traceback.format_exc()
        return {
            "ERROR": f"Error getting stream previews from source '{source_name}': "
            f"{ex!r}, {ex!s}\n{tb_str}"
        }

    result: dict[str, list[dict[str, Any]] | str] = {}
    for stream_name, dataset in samples_result.items():
        if dataset is None:
            result[stream_name] = f"Could not retrieve stream samples for stream '{stream_name}'"
        else:
            result[stream_name] = list(dataset)

    return result


@mcp_tool(
    destructive=False,
    extra_help_text=_CONFIG_HELP,
)
def sync_source_to_cache(
    source_connector_name: Annotated[
        str,
        Field(description="The name of the source connector."),
    ],
    config: Annotated[
        dict | str | None,
        Field(
            description="The configuration for the source connector as a dict or JSON string.",
            default=None,
        ),
    ],
    config_file: Annotated[
        str | Path | None,
        Field(
            description="Path to a YAML or JSON file containing the source connector config.",
            default=None,
        ),
    ],
    config_secret_name: Annotated[
        str | None,
        Field(
            description="The name of the secret containing the configuration.",
            default=None,
        ),
    ],
    streams: Annotated[
        list[str] | str,
        Field(
            description="The streams to sync.",
            default="suggested",
        ),
    ],
    override_execution_mode: Annotated[
        Literal["docker", "python", "yaml", "auto"],
        Field(
            description="Optionally override the execution method to use for the connector. "
            "This parameter is ignored if manifest_path is provided (yaml mode will be used).",
            default="auto",
        ),
    ],
    manifest_path: Annotated[
        str | Path | None,
        Field(
            description="Path to a local YAML manifest file for declarative connectors.",
            default=None,
        ),
    ],
) -> str:
    """Run a sync from a source connector to the default DuckDB cache."""
    source: Source = _get_mcp_source(
        connector_name=source_connector_name,
        override_execution_mode=override_execution_mode,
        manifest_path=manifest_path,
    )
    config_dict = resolve_connector_config(
        config=config,
        config_file=config_file,
        config_secret_name=config_secret_name,
        config_spec_jsonschema=source.config_spec,
    )
    source.set_config(config_dict)
    cache = get_default_cache()

    streams = resolve_list_of_strings(streams)
    if streams and len(streams) == 1 and streams[0] in {"*", "suggested"}:
        # Float '*' and 'suggested' to the top-level for special processing:
        streams = streams[0]

    if isinstance(streams, str) and streams == "suggested":
        streams = "*"  # Default to all streams if 'suggested' is not otherwise specified.
        try:
            metadata = get_connector_metadata(
                source_connector_name,
            )
        except Exception:
            streams = "*"  # Fallback to all streams if suggested streams fail.
        else:
            if metadata is not None:
                streams = metadata.suggested_streams or "*"

    if isinstance(streams, str) and streams != "*":
        streams = [streams]  # Ensure streams is a list

    source.read(
        cache=cache,
        streams=streams,
    )
    del cache  # Ensure the cache is closed properly

    summary: str = f"Sync completed for '{source_connector_name}'!\n\n"
    summary += "Data written to default DuckDB cache\n"
    return summary


class CachedDatasetInfo(BaseModel):
    """Class to hold information about a cached dataset."""

    stream_name: str
    """The name of the stream in the cache."""
    table_name: str
    schema_name: str | None = None


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def list_cached_streams() -> list[CachedDatasetInfo]:
    """List all streams available in the default DuckDB cache."""
    cache: DuckDBCache = get_default_cache()
    result = [
        CachedDatasetInfo(
            stream_name=stream_name,
            table_name=(cache.table_prefix or "") + stream_name,
            schema_name=cache.schema_name,
        )
        for stream_name in cache.streams
    ]
    del cache  # Ensure the cache is closed properly
    return result


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def describe_default_cache() -> dict[str, Any]:
    """Describe the currently configured default cache."""
    cache = get_default_cache()
    return {
        "cache_type": type(cache).__name__,
        "cache_dir": str(cache.cache_dir),
        "cache_db_path": str(Path(cache.db_path).absolute()),
        "cached_streams": list(cache.streams.keys()),
    }


def _is_safe_sql(sql_query: str) -> bool:
    """Check if a SQL query is safe to execute.

    For security reasons, we only allow read-only operations like SELECT, DESCRIBE, and SHOW.
    Multi-statement queries (containing semicolons) are also disallowed for security.

    Note: SQLAlchemy will also validate downstream, but this is a first-pass check.

    Args:
        sql_query: The SQL query to check

    Returns:
        True if the query is safe to execute, False otherwise
    """
    # Remove leading/trailing whitespace and convert to uppercase for checking
    normalized_query = sql_query.strip().upper()

    # Disallow multi-statement queries (containing semicolons)
    # Note: We check the original query to catch semicolons anywhere, including in comments
    if ";" in sql_query:
        return False

    # List of allowed SQL statement prefixes (read-only operations)
    allowed_prefixes = (
        "SELECT",
        "DESCRIBE",
        "DESC",  # Short form of DESCRIBE
        "SHOW",
        "EXPLAIN",  # Also safe - shows query execution plan
    )

    # Check if the query starts with any allowed prefix
    return any(normalized_query.startswith(prefix) for prefix in allowed_prefixes)


@mcp_tool(
    read_only=True,
    idempotent=True,
    extra_help_text=_CONFIG_HELP,
)
def run_sql_query(
    sql_query: Annotated[
        str,
        Field(description="The SQL query to execute."),
    ],
    max_records: Annotated[
        int,
        Field(
            description="Maximum number of records to return.",
            default=1000,
        ),
    ],
) -> list[dict[str, Any]]:
    """Run a SQL query against the default cache.

    The dialect of SQL should match the dialect of the default cache.
    Use `describe_default_cache` to see the cache type.

    For DuckDB-type caches:
    - Use `SHOW TABLES` to list all tables.
    - Use `DESCRIBE <table_name>` to get the schema of a specific table

    For security reasons, only read-only operations are allowed: SELECT, DESCRIBE, SHOW, EXPLAIN.
    """
    # Check if the query is safe to execute
    if not _is_safe_sql(sql_query):
        return [
            {
                "ERROR": "Unsafe SQL query detected. Only read-only operations are allowed: "
                "SELECT, DESCRIBE, SHOW, EXPLAIN",
                "SQL_QUERY": sql_query,
            }
        ]

    cache: DuckDBCache = get_default_cache()
    try:
        return cache.run_sql_query(
            sql_query,
            max_records=max_records,
        )
    except Exception as ex:
        tb_str = traceback.format_exc()
        return [
            {
                "ERROR": f"Error running SQL query: {ex!r}, {ex!s}",
                "TRACEBACK": tb_str,
                "SQL_QUERY": sql_query,
            }
        ]
    finally:
        del cache  # Ensure the cache is closed properly


def register_local_tools(app: FastMCP) -> None:
    """Register local tools with the FastMCP app.

    Args:
        app: FastMCP application instance
    """
    register_mcp_tools(app, mcp_module=__name__)
